from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

ChildProtectionPlans = CINTable.ChildProtectionPlans
CINdetails = CINTable.CINdetails
Section47 = CINTable.Section47
CINplanDates = CINTable.CINplanDates

ReasonForClosure = CINdetails.ReasonForClosure
LAchildID = ChildProtectionPlans.LAchildID
CINdetailsID = ChildProtectionPlans.CINdetailsID
DateOfInitialCPC = CINdetails.DateOfInitialCPC


# define characteristics of rule
@rule_definition(
    code="2990",
    module=CINTable.CINdetails,
    message="Activity is recorded against a case marked as ‘Case closed after assessment, no further action’ or 'case closed after assessment, referred to early help'.",
    affected_fields=[
        ReasonForClosure,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_cpp = data_container[ChildProtectionPlans].copy()
    df_47 = data_container[Section47].copy()
    df_cin = data_container[CINdetails].copy()
    df_cin_pd = data_container[CINplanDates].copy()

    df_cpp.index.name = "ROW_ID"
    df_47.index.name = "ROW_ID"
    df_cin.index.name = "ROW_ID"
    df_cin_pd.index.name = "ROW_ID"

    df_cpp.reset_index(inplace=True)
    df_47.reset_index(inplace=True)
    df_cin.reset_index(inplace=True)
    df_cin_pd.reset_index(inplace=True)

    # If a <CINDetails> module has <ReasonForClosure> (N00103) = RC8 or RC9, then it cannot have any of the following modules:
    # <Section47> module
    # <ChildProtectionPlan> module
    # <DateofInitialCPC> (N00110) within the <CINDetails> module
    # <CINPlanDates> module
    df_cin = df_cin[df_cin[ReasonForClosure].isin(["RC8", "RC9"])]

    df_cin_cpp = df_cin.merge(
        df_cpp, on=["LAchildID", "CINdetailsID"], how="left", suffixes=["_cin", "_cpp"]
    )

    df_cin_47 = df_cin.merge(
        df_47,
        on=[
            "LAchildID",
            "CINdetailsID",
        ],
        how="left",
        suffixes=["_cin", "_47"],
    )

    df_cin_cin_pd = df_cin.merge(
        df_cin_pd,
        on=["LAchildID", "CINdetailsID"],
        how="left",
        suffixes=["_cin", "_pd"],
    )

    df_cin_cpp_47 = df_cin_cpp.merge(
        df_cin_47,
        left_on=[
            "LAchildID",
            "CINdetailsID",
            "ROW_ID_cin",
            "DateOfInitialCPC",
            "ReasonForClosure",
        ],
        right_on=[
            "LAchildID",
            "CINdetailsID",
            "ROW_ID_cin",
            "DateOfInitialCPC_cin",
            "ReasonForClosure",
        ],
        how="left",
        suffixes=["_cin_cpp", "_cin_47"],
    )

    # This merge uses the DateOfInitialCPC values from the original CIN table using their different suffixes from each merge
    merged_df = df_cin_cpp_47.merge(
        df_cin_cin_pd,
        left_on=[
            "LAchildID",
            "CINdetailsID",
            "ROW_ID_cin",
            "DateOfInitialCPC_cin",
            "ReasonForClosure",
        ],
        right_on=[
            "LAchildID",
            "CINdetailsID",
            "ROW_ID_cin",
            "DateOfInitialCPC",
            "ReasonForClosure",
        ],
        how="left",
        suffixes=["_cin_cpp_47", "_cin_cin_pd"],
    )

    # Logical conditions - other than this, of the tables can merge, it means there's modules and they are in error
    # Checks for the DateOfInitialCPC caqrried on through from the original CINdetails table
    condition_1 = merged_df["DateOfInitialCPC_cin"].notna()
    condition_2 = merged_df["ROW_ID_cpp"].notna()
    condition_3 = merged_df["ROW_ID_47"].notna()
    condition_4 = merged_df["ROW_ID_pd"].notna()

    merged_df = merged_df[
        condition_1 | condition_2 | condition_3 | condition_4
    ].reset_index()

    merged_df["ERROR_ID"] = tuple(zip(merged_df[LAchildID], merged_df[CINdetailsID]))

    df_cpp_issues = (
        df_cpp.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_cpp")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_47_issues = (
        df_47.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_47")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_cin_issues = (
        df_cin.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    df_cin_pd_issues = (
        df_cin_pd.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_pd")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=ChildProtectionPlans, columns=[LAchildID], row_df=df_cpp_issues
    )
    rule_context.push_type_2(
        table=CINdetails, columns=[DateOfInitialCPC], row_df=df_cin_issues
    )
    rule_context.push_type_2(
        table=CINplanDates, columns=[LAchildID], row_df=df_cin_pd_issues
    )
    rule_context.push_type_2(table=Section47, columns=[LAchildID], row_df=df_47_issues)


def test_validate():
    sample_cpp = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CPPstartDate": "26/05/2000",
                "CINdetailsID": "cinID10",
            },
            {
                "LAchildID": "child1",
                "CPPstartDate": "27/06/2002",
                "CINdetailsID": "cinID2",
            },
            {  # fails for having a module
                "LAchildID": "child3",
                "CPPstartDate": "26/05/2000",
                "CINdetailsID": "cinID3",
            },
            {
                "LAchildID": "child30",
                "CPPstartDate": "26/05/2000",
                "CINdetailsID": "cinID3",
            },
            {
                "LAchildID": "child30",
                "CPPstartDate": pd.NA,
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child3",
                "CPPstartDate": "07/02/2001",
                "CINdetailsID": "cinID30",
            },
            {
                "LAchildID": "child3",
                "CPPstartDate": "14/03/2001",
                "CINdetailsID": "cinID4",
            },
        ]
    )
    sample_section47 = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child2",  # fails for having a module
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child2",
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2000",
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID3",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": "14/03/2001",
                "CINdetailsID": "cinID4",
            },
        ]
    )
    sample_cin_details = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "DateOfInitialCPC": pd.NA,  # 0 fail: found in CINplanDates table
                "CINdetailsID": "cinID1",
                "ReasonForClosure": "RC8",
            },
            {
                "LAchildID": "child2",
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID2",  # 1 fail: found in Section47 table
                "ReasonForClosure": "RC9",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": pd.NA,  # 2 fail : found in Section47 and CPP tables
                "CINdetailsID": "cinID3",
                "ReasonForClosure": "RC8",
            },
            {
                "LAchildID": "child4",
                "DateOfInitialCPC": "28/05/2000",  # 3 fails for having initialcpc
                "CINdetailsID": "cinID4",
                "ReasonForClosure": "RC8",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2000",
                "CINdetailsID": "cinID2",
                "ReasonForClosure": "RC10",  # 4 ignore: reason != RC8
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2003",
                "CINdetailsID": "cinID8",
                "ReasonForClosure": "RC10",  # 5 ignore: reason != RC8
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": "14/03/2001",
                "CINdetailsID": "cinID4",
                "ReasonForClosure": "RC10",  # 6 ignore: reason != RC8
            },
            {
                "LAchildID": "child7",
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID4",
                "ReasonForClosure": "RC10",  # 7 pass
            },
        ]
    )
    sample_cin_plan_dates = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # fails for having a module
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "cinID3",
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "cinID4",
            },
        ]
    )

    sample_cin_details["DateOfInitialCPC"] = pd.to_datetime(
        sample_cin_details["DateOfInitialCPC"], format="%d/%m/%Y", errors="coerce"
    )
    sample_section47["DateOfInitialCPC"] = pd.to_datetime(
        sample_section47["DateOfInitialCPC"], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(
        validate,
        {
            ChildProtectionPlans: sample_cpp,
            Section47: sample_section47,
            CINdetails: sample_cin_details,
            CINplanDates: sample_cin_plan_dates,
        },
    )

    issues_list = result.type2_issues
    assert len(issues_list) == 4
    issues = issues_list[1]

    issue_table = issues.table
    assert issue_table == CINdetails

    issue_columns = issues.columns
    assert issue_columns == [DateOfInitialCPC]

    issue_rows = issues.row_df
    assert len(issue_rows) == 4
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",  # ChildID
                    "cinID1",  # CINdetailsID
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child2",  # ChildID
                    "cinID2",  # CINdetailsID
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child3",  # ChildID
                    "cinID3",  # CINdetailsID
                ),
                "ROW_ID": [2],
            },
            {
                "ERROR_ID": (
                    "child4",  # ChildID
                    "cinID4",  # CINdetailsID
                ),
                "ROW_ID": [3],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "2990"
    assert (
        result.definition.message
        == "Activity is recorded against a case marked as ‘Case closed after assessment, no further action’ or 'case closed after assessment, referred to early help'."
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

CINDetails = CINTable.CINdetails
CINclosureDate = CINDetails.CINclosureDate
ReasonForClosure = CINDetails.ReasonForClosure
LAchildID = CINDetails.LAchildID


# define characteristics of rule
@rule_definition(
    # write the rule code here, in place of '8805'
    code="8805",
    module=CINTable.CINdetails,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="A CIN case cannot have a CIN closure date without a Reason for Closure",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[CINclosureDate, ReasonForClosure],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    df = data_container[CINDetails]
    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df.index.name = "ROW_ID"

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # If <CINclosureDate> (N00102) is present then <ReasonForClosure> (N00103) must also be present
    # return rows where CINClosureDate is present but ReasonForClosure is not.
    # If CINclosureDate is not null and ReasonForClosure is null
    condition = df[CINclosureDate].notna() & df[ReasonForClosure].isna()

    # get all the data that fits the failing condition. Reset the index so that ROW_ID now becomes a column of df
    df_issues = df[condition].reset_index()

    # SUBMIT ERRORS

    link_id = tuple(
        zip(
            df_issues[LAchildID], df_issues[CINclosureDate], df_issues[ReasonForClosure]
        )
    )
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    # Ensure that you do not change the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_1(
        table=CINDetails, columns=[CINclosureDate, ReasonForClosure], row_df=df_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    fake_date_frame = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINclosureDate": "26/05/2000",
                "ReasonForClosure": "26/05/2000",
            },
            {
                "LAchildID": "child2",
                "CINclosureDate": pd.NA,
                "ReasonForClosure": "26/05/2000",
            },
            {
                "LAchildID": "child3",
                "CINclosureDate": "26/05/1999",
                "ReasonForClosure": "26/05/2000",
            },
            {
                "LAchildID": "child3",
                "CINclosureDate": "26/05/2000",
                "ReasonForClosure": pd.NA,
            },  # fail because CINClosureDate is populated and ReasonForClosure isn't
            {
                "LAchildID": "child4",
                "CINclosureDate": "25/05/2000",
                "ReasonForClosure": pd.NA,
            },  # fail because CINClosureDate is populated and ReasonForClosure isn't
            {
                "LAchildID": "child5",
                "CINclosureDate": pd.NA,
                "ReasonForClosure": pd.NA,
            },
        ]
    )

    # Date values not checked so no datetime conversion required

    # Run rule function passing in our sample data
    result = run_rule(validate, {CINDetails: fake_date_frame})

    # Use .type1_issues to check for the result of .push_type1_issues() which you used above.
    issues = result.type1_issues

    # get table name and check it.
    issue_table = issues.table
    assert issue_table == CINDetails

    # check that the right columns were returned.
    issue_columns = issues.columns
    assert issue_columns == [CINclosureDate, ReasonForClosure]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 2
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child3",
                    "26/05/2000",
                    pd.NA,
                ),
                "ROW_ID": [3],
            },
            {
                "ERROR_ID": (
                    "child4",
                    "25/05/2000",
                    pd.NA,
                ),
                "ROW_ID": [4],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace '8805' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8805"
    assert (
        result.definition.message
        == "A CIN case cannot have a CIN closure date without a Reason for Closure"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

CINdetails = CINTable.CINdetails
LAchildID = CINdetails.LAchildID
CINdetailsID = CINdetails.CINdetailsID
ReferralNFA = CINdetails.ReferralNFA
CINreferralDate = CINdetails.CINreferralDate
CINclosureDate = CINdetails.CINclosureDate


# define characteristics of rule
@rule_definition(
    # write the rule code here
    code="8816",
    # replace CINdetails with the value in the module column of the excel sheet corresponding to this rule .
    # Note that even if multiple tables are involved, one table will be named in the module column.
    module=CINTable.CINdetails,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="An open CIN episode is shown and case is not a referral with no further action, but it is not the latest episode.",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[
        CINreferralDate,
        ReferralNFA,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    df_cin = data_container[CINdetails].copy()

    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df_cin.index.name = "ROW_ID"

    # If there is a <CINdetails> module present for the child where:
    # <CINclosureDate> (N00102) is missing; and
    # <ReferralNFA> (N00112) = false or 0
    # then the <CINreferralDate> (N00100) for this module must be the latest of all Referral Dates for that child.

    # removing nans prevents ValueError: attempt to get argmax of an empty sequence, when no CINreferralDate is present
    df_cin = df_cin[df_cin[CINreferralDate].notna()]

    # find out the latest referral date (and its index position) for each child and attach it to all the rows for that child.
    df_cin["latest_referral"] = df_cin.groupby(LAchildID)[CINreferralDate].transform(
        "max"
    )
    df_cin["latest_referral_ind"] = df_cin.groupby(LAchildID)[
        CINreferralDate
    ].transform("idxmax")

    # resetting the index makes ROW_ID into a column to keep it intact. The logic can now begin.
    df_cin.reset_index(inplace=True)

    # filter the rows of interest

    df_cin = df_cin[
        (df_cin[CINclosureDate].isna()) & (df_cin[ReferralNFA].isin(["false", "0"]))
    ]
    # check those that do not meet the requirements
    condition = df_cin["ROW_ID"] != df_cin["latest_referral_ind"]
    df_cin = df_cin[condition]

    df_cin["ERROR_ID"] = tuple(
        zip(df_cin[LAchildID], df_cin[CINdetailsID], df_cin[CINreferralDate])
    )

    df_cin_issues = (
        df_cin.groupby("ERROR_ID", group_keys=False)["ROW_ID"].apply(list).reset_index()
    )
    # the locations of the latest_referral date should also be flagged
    df_cin_issues_cause = (
        df_cin.groupby("ERROR_ID", group_keys=False)["latest_referral_ind"]
        .apply(list)
        .reset_index()
    )
    # rename column to fit expected structure of a df_issues dataframe
    df_cin_issues_cause.rename(columns={"latest_referral_ind": "ROW_ID"}, inplace=True)

    # Ensure that you maintain the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_3(
        table=CINdetails,
        columns=[CINreferralDate],
        row_df=df_cin_issues,
    )
    rule_context.push_type_3(
        table=CINdetails,
        columns=[CINreferralDate],
        row_df=df_cin_issues_cause,
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.

    sample_cin = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINclosureDate": pd.NA,
                "CINreferralDate": "26/10/2000",  # 0 fail: not latest referral in child1
                "CINdetailsID": "cinID1",
                "ReferralNFA": "0",
            },
            {
                "LAchildID": "child1",
                "CINclosureDate": "26/05/2001",
                "CINreferralDate": "26/10/2001",  # 1 ignore: CINclosureDate notna
                "CINdetailsID": "cinID2",
                "ReferralNFA": "0",
            },
            {
                "LAchildID": "child1",
                "CINclosureDate": pd.NA,
                "CINreferralDate": "26/10/1999",  # 2 fail: not latest referral in child1
                "CINdetailsID": "cinID3",
                "ReferralNFA": "0",
            },
            {
                "LAchildID": "child1",
                "CINclosureDate": pd.NA,
                "CINreferralDate": "26/10/2002",  # 3 pass: latest referral in child1
                "CINdetailsID": "cinID4",
                "ReferralNFA": "0",
            },
            # child2
            {
                "LAchildID": "child2",
                "CINclosureDate": pd.NA,
                "CINreferralDate": "26/10/1999",  # 4 pass: latest referral in child2
                "CINdetailsID": "cinID3",
                "ReferralNFA": "0",
            },
            {
                "LAchildID": "child2",
                "CINclosureDate": pd.NA,
                "CINreferralDate": "26/10/1990",  # 5 fail: not latest referral in child2
                "CINdetailsID": "cinID3",
                "ReferralNFA": "0",
            },
            {
                "LAchildID": "child4",
                "CINclosureDate": pd.NA,
                "CINreferralDate": pd.NA,
                "CINdetailsID": "cinID3",
                "ReferralNFA": "0",
            },
            {
                "LAchildID": "child4",
                "CINclosureDate": pd.NA,
                "CINreferralDate": pd.NA,
                "CINdetailsID": "cinID4",
                "ReferralNFA": "0",
            },
            {
                "LAchildID": "child5",
                "CINreferralDate": "26/10/1999",
                "CINdetailsID": "cinID4",
                "ReferralNFA": "0",
            },
            {
                "LAchildID": "child5",  # 9, Fail, shared CINreferralDate
                "CINdetailsID": "cinID5",
                "CINreferralDate": "26/10/1999",
                "ReferralNFA": "0",
            },
        ]
    )

    # If rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    sample_cin[CINclosureDate] = pd.to_datetime(
        sample_cin[CINclosureDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_cin[CINreferralDate] = pd.to_datetime(
        sample_cin[CINreferralDate], format="%d/%m/%Y", errors="coerce"
    )

    # Run the rule function, passing in our sample data.
    result = run_rule(
        validate,
        {
            CINdetails: sample_cin,
        },
    )

    # Use .type3_issues to check for the result of .push_type2_issues() which you used above.
    issues_list = result.type3_issues
    assert len(issues_list) == 2
    # the function returns a list on NamedTuples where each NamedTuple contains (table, column_list, df_issues)
    # pick any table and check it's values.
    issues = issues_list[0]

    issue_table = issues.table
    assert issue_table == CINdetails

    issue_columns = issues.columns
    assert issue_columns == [CINreferralDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 3 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 4

    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on, in your zip, earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    "cinID1",
                    pd.to_datetime("26/10/2000", format="%d/%m/%Y"),
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child1",
                    "cinID3",
                    pd.to_datetime("26/10/1999", format="%d/%m/%Y"),
                ),
                "ROW_ID": [2],
            },
            {
                "ERROR_ID": (
                    "child2",
                    "cinID3",
                    pd.to_datetime("26/10/1990", format="%d/%m/%Y"),
                ),
                "ROW_ID": [5],
            },
            {
                "ERROR_ID": (
                    "child5",
                    "cinID5",
                    pd.to_datetime("26/10/1999", format="%d/%m/%Y"),
                ),
                "ROW_ID": [9],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace '8816' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8816"
    assert (
        result.definition.message
        == "An open CIN episode is shown and case is not a referral with no further action, but it is not the latest episode."
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

Section47 = CINTable.Section47
LAchildID = Section47.LAchildID
CINdetailsID = Section47.CINdetailsID
DateOfInitialCPC = Section47.DateOfInitialCPC
ICPCnotRequired = Section47.ICPCnotRequired


# define characteristics of rule
@rule_definition(
    code="8839",
    module=CINTable.Section47,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="Within one CINDetails group there are 2 or more open S47 Assessments",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[DateOfInitialCPC],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    # Replace Section47 with the name of the table you need.
    df = data_container[Section47]
    # Before you begin, rename the index and make it a column, so that the initial row positions can be kept intact.
    df.index.name = "ROW_ID"
    df.reset_index(inplace=True)

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # Within one <CINdetails> group, there must be only one <Section47> group that has no <DateOfInitialCPC> (N00110) recorded
    # OR <Section47> group that has a missing <DateOfInitialCPC> (N00110) and the <ICPCnotRequired> (N00111) flag is not true

    # DF_CHECK: APPLY GROUPBYs IN A SEPARATE DATAFRAME SO THAT OTHER COLUMNS ARE NOT LOST OR CORRUPTED. THEN, MAP THE RESULTS TO THE INITIAL DATAFRAME.
    df_check = df.copy()
    # get all the locations where ICPCnotRequired is null or not true (1)
    # The rule originally asks for true, not 1, but an analyst coded it for 1, so their LA may use 1 and 0 instead, as such, it now checks for either.
    df_check = df_check[
        df_check[ICPCnotRequired].isna()
        | (~df_check[ICPCnotRequired].astype(str).isin(["true", "1"]))
    ]
    # get all the locations where DateOfInitialCPC is null
    df_check = df_check[df_check[DateOfInitialCPC].isna()]
    # We'll have to count the number of nan values per group. NaNs cannot be counted so replace them with something that can.
    # Do this only if your rule requires that you interact with a column made up of all NaNs.
    # Adding as a separate column because I need the original column in the groupby for df_issues later on.
    df_check["CountICPC"] = df_check[DateOfInitialCPC]
    df_check["CountICPC"].fillna(1, inplace=True)
    # count how many occurences of missing DateOfInitialCPC per CINdetails group in each child.
    df_check = (
        df_check.groupby([LAchildID, CINdetailsID, DateOfInitialCPC], dropna=False)[
            "CountICPC"
        ]
        .count()
        .reset_index()
    )

    # when you groupby as shown above a series is returned where the columns in the round brackets become the index and the groupby result are the values.
    # resetting the index pushes the columns in the () back as columns of the dataframe and assigns the groupby result to the column in the square bracket.

    # filter out the instances where DateOfInitialCPC is missing more than once in a CINdetails group.
    df_check = df_check[df_check["CountICPC"] > 1]
    issue_ids = tuple(
        zip(df_check[LAchildID], df_check[CINdetailsID], df_check[DateOfInitialCPC])
    )
    # DF_ISSUES: GET ALL THE DATA ABOUT THE LOCATIONS THAT WERE IDENTIFIED IN DF_CHECK
    df["ERROR_ID"] = tuple(zip(df[LAchildID], df[CINdetailsID], df[DateOfInitialCPC]))
    df_issues = df[df.ERROR_ID.isin(issue_ids)]
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # Ensure that you do not change the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_3(
        table=Section47, columns=[DateOfInitialCPC], row_df=df_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_s47 = pd.DataFrame(
        [  # child1
            {  # fail
                LAchildID: "child1",
                CINdetailsID: "cinID1",
                DateOfInitialCPC: pd.NA,  # 0 first nan date in group
                ICPCnotRequired: 0,
            },
            {  # fail
                LAchildID: "child1",
                CINdetailsID: "cinID1",
                DateOfInitialCPC: pd.NA,  # 1 second nan date in group
                ICPCnotRequired: 0,
            },
            {  # pass
                LAchildID: "child1",
                CINdetailsID: "cinID2",
                DateOfInitialCPC: pd.NA,  # 2 not more than one nan authorisation date in group
                ICPCnotRequired: 0,
            },
            # child2
            {  # pass
                LAchildID: "child2",
                CINdetailsID: "cinID2",
                DateOfInitialCPC: "26/05/2021",  # 3 not nan
                ICPCnotRequired: 0,
            },
            {  # fail
                LAchildID: "child2",
                CINdetailsID: "cinID2",
                DateOfInitialCPC: pd.NA,  # 4 first nan date in group
                ICPCnotRequired: pd.NA,
            },
            {  # fail
                LAchildID: "child2",
                CINdetailsID: "cinID2",
                DateOfInitialCPC: pd.NA,  # 5 second nan date in group
                ICPCnotRequired: pd.NA,
            },
            # child 3
            {  # pass
                LAchildID: "child3",
                CINdetailsID: "cinID3",
                DateOfInitialCPC: pd.NA,  # 6 nan but also ICPC not required
                ICPCnotRequired: "true",
            },
            {  # pass
                LAchildID: "child3",
                CINdetailsID: "cinID3",
                DateOfInitialCPC: pd.NA,  # 7 not more than one nan authorisation date in group
                ICPCnotRequired: pd.NA,
            },
            {  # pass
                LAchildID: "child3",
                CINdetailsID: "cinID3",
                DateOfInitialCPC: pd.NA,  # 6 nan but also ICPC not required
                ICPCnotRequired: 1,
            },
            {  # pass, 1 as string for issue 373
                LAchildID: "child3",
                CINdetailsID: "cinID3",
                DateOfInitialCPC: pd.NA,
                ICPCnotRequired: "1",
            },
        ]
    )
    # if rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    sample_s47[DateOfInitialCPC] = pd.to_datetime(
        sample_s47[DateOfInitialCPC],
        format="%d/%m/%Y",
        errors="coerce",
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {Section47: sample_s47})

    # Use .type3_issues to check for the result of .push_type3_issues() which you used above.
    issues_list = result.type3_issues
    # Issues list contains the objects pushed in their respective order. Since push_type3 was only used once, there will be one object in issues_list.
    assert len(issues_list) == 1

    issues = issues_list[0]

    issue_table = issues.table
    assert issue_table == Section47

    issue_columns = issues.columns
    assert issue_columns == [DateOfInitialCPC]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 2
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    "cinID1",
                    pd.NaT,
                ),
                "ROW_ID": [0, 1],
            },
            {
                "ERROR_ID": (
                    "child2",
                    "cinID2",
                    pd.NaT,
                ),
                "ROW_ID": [4, 5],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace '8839' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8839"
    assert (
        result.definition.message
        == "Within one CINDetails group there are 2 or more open S47 Assessments"
    )



"""
Rule number: '8606'
Module: CIN details
Rule details: <CINreferralDate> (N00100) cannot be more than 280 days before <PersonBirthDate> (N00066) or <ExpectedPersonBirthDate> (N00098)
Rule message: Child referral date is more than 40 weeks before DOB or expected DOB

"""

import datetime as dt
from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
PersonBirthDate = ChildIdentifiers.PersonBirthDate
ExpectedPersonBirthDate = ChildIdentifiers.ExpectedPersonBirthDate
LAchildID = ChildIdentifiers.LAchildID

CINdetails = CINTable.CINdetails
CINreferralDate = CINdetails.CINreferralDate
LAchildID = CINdetails.LAchildID


@rule_definition(
    code="8606",
    module=CINTable.CINdetails,
    message="Child referral date is more than 40 weeks before DOB or expected DOB",
    affected_fields=[
        CINreferralDate,
        PersonBirthDate,
        ExpectedPersonBirthDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_CINDetails = data_container[CINdetails].copy()
    df_ChildIdentifiers = data_container[ChildIdentifiers].copy()

    df_CINDetails.index.name = "ROW_ID"
    df_ChildIdentifiers.index.name = "ROW_ID"

    df_CINDetails.reset_index(inplace=True)
    df_ChildIdentifiers.reset_index(inplace=True)

    # <CINreferralDate> (N00100) cannot be more than 280 days before <PersonBirthDate> (N00066) or <ExpectedPersonBirthDate>
    df_merged = df_CINDetails.merge(
        df_ChildIdentifiers,
        left_on=["LAchildID"],
        right_on=["LAchildID"],
        how="left",
        suffixes=("_CINDetails", "_ChildIdentifiers"),
    )

    # Get rows where CINreferralDate is earlier than birth/expected birth -280
    condition1 = df_merged[CINreferralDate] < (
        df_merged[PersonBirthDate] - dt.timedelta(days=280)
    )
    condition2 = df_merged[CINreferralDate] < (
        df_merged[ExpectedPersonBirthDate] - dt.timedelta(days=280)
    )
    df_merged = df_merged[condition1 | condition2].reset_index()

    df_merged["ERROR_ID"] = tuple(
        zip(
            df_merged[LAchildID],
            df_merged[CINreferralDate],
        )
    )
    df_CINDetails_issues = (
        df_CINDetails.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_CINDetails")
        .groupby("ERROR_ID", group_keys="False")["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_ChildIdentifiers_issues = (
        df_ChildIdentifiers.merge(
            df_merged, left_on="ROW_ID", right_on="ROW_ID_ChildIdentifiers"
        )
        .groupby("ERROR_ID", group_keys="False")["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=CINdetails, columns=[CINreferralDate], row_df=df_CINDetails_issues
    )
    rule_context.push_type_2(
        table=ChildIdentifiers,
        columns=[PersonBirthDate, ExpectedPersonBirthDate],
        row_df=df_ChildIdentifiers_issues,
    )


def test_validate():
    sample_CINDetails = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINreferralDate": "26/04/2000",  # Pass birth less than 280 days before referral
            },
            {
                "LAchildID": "child2",
                "CINreferralDate": "27/06/1998",  # Fail, referral more than 280 days before birth
            },
            {
                "LAchildID": "child3",
                "CINreferralDate": "07/04/2000",  # Pass, expected birth less than 280 days before referral
            },
            {
                "LAchildID": "child4",
                "CINreferralDate": "07/02/1998",  # Fail, referral date more than 280 days before expected birth
            },
        ]
    )
    sample_ChildIdentifiers = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # Pass
                "PersonBirthDate": "26/05/2000",
                "ExpectedPersonBirthDate": pd.NA,
            },
            {
                "LAchildID": "child2",  # Fails
                "PersonBirthDate": "26/05/2000",
                "ExpectedPersonBirthDate": pd.NA,
            },
            {
                "LAchildID": "child3",  # Pass
                "PersonBirthDate": pd.NA,
                "ExpectedPersonBirthDate": "26/05/2000",
            },
            {
                "LAchildID": "child4",  # Fail
                "PersonBirthDate": pd.NA,
                "ExpectedPersonBirthDate": "26/05/2000",
            },
        ]
    )

    sample_CINDetails[CINreferralDate] = pd.to_datetime(
        sample_CINDetails[CINreferralDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_ChildIdentifiers["PersonBirthDate"] = pd.to_datetime(
        sample_ChildIdentifiers["PersonBirthDate"], format="%d/%m/%Y", errors="coerce"
    )

    sample_ChildIdentifiers["ExpectedPersonBirthDate"] = pd.to_datetime(
        sample_ChildIdentifiers["ExpectedPersonBirthDate"],
        format="%d/%m/%Y",
        errors="coerce",
    )

    result = run_rule(
        validate,
        {
            CINdetails: sample_CINDetails,
            ChildIdentifiers: sample_ChildIdentifiers,
        },
    )

    issues_list = result.type2_issues
    assert len(issues_list) == 2
    issues = issues_list[1]

    issue_table = issues.table

    assert issue_table == ChildIdentifiers

    issue_columns = issues.columns

    assert issue_columns == [PersonBirthDate, ExpectedPersonBirthDate]

    issue_rows = issues.row_df

    assert len(issue_rows) == 2
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child2",  # ChildID
                    # Referral date
                    pd.to_datetime("27/06/1998", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child4",  # ChildID
                    # Referral date
                    pd.to_datetime("07/02/1998", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [3],
            },
        ]
    )

    assert issue_rows.equals(expected_df)

    assert result.definition.code == "8606"
    assert (
        result.definition.message
        == "Child referral date is more than 40 weeks before DOB or expected DOB"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.test_engine import run_rule
from cin_validator.utils import england_working_days, make_census_period

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

Section47 = CINTable.Section47
DateOfInitialCPC = Section47.DateOfInitialCPC
ICPCnotReqiured = Section47.ICPCnotRequired
S47ActualStartDate = Section47.S47ActualStartDate
LAchildID = Section47.LAchildID

Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


# define characteristics of rule
@rule_definition(
    # write the rule code here, in place of 8675Q
    code="8675Q",
    rule_type=RuleType.QUERY,
    module=CINTable.Section47,
    message="Please check and either amend data or provide a reason: S47 Enquiry started more than 15 working days before the end of the census year. However, there is no date of Initial Child Protection Conference.",
    affected_fields=[DateOfInitialCPC, S47ActualStartDate, ICPCnotReqiured],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA
    df = data_container[Section47]
    header = data_container[Header]
    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df.index.name = "ROW_ID"

    ref_date_series = header[ReferenceDate]
    collection_start, collection_end = make_census_period(ref_date_series)

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # If <DateOfInitialCPC> (N00110) not present and <ICPCnotReqiured> (N00111) equals false
    # then <S47ActualStartDate> (N00148) should not be before the <ReferenceDate> (N00603) minus 15 working days
    no_cpc = df[DateOfInitialCPC].isna()
    icpc_false = df[ICPCnotReqiured].astype(str).isin(["false", "0"])
    before_15b = df[S47ActualStartDate] < (collection_end - england_working_days(15))
    condition = (no_cpc & icpc_false) & (before_15b)

    # get all the data that fits the failing condition. Reset the index so that ROW_ID now becomes a column of df
    df_issues = df[condition].reset_index()

    # SUBMIT ERRORS
    # Generate a unique ID for each instance of an error.
    # Replace S47ActualStartDate and ICPCnotReqiured below with the columns concerned in your rule.
    link_id = tuple(
        zip(
            df_issues[LAchildID],
            df_issues[S47ActualStartDate],
            df_issues[ICPCnotReqiured],
        )
    )
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    # Ensure that you do not change the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_1(
        table=Section47,
        columns=[DateOfInitialCPC, S47ActualStartDate, ICPCnotReqiured],
        row_df=df_issues,
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    fake_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2022"}]  # the census start date here will be 01/04/2021
    )

    section47 = pd.DataFrame(
        [
            {  # 0 fail, no ICPCnotrequied as true or InitialCPC, and date is more than 15 days before end of census year
                "LAchildID": "child1",
                "DateOfInitialCPC": pd.NA,
                "S47ActualStartDate": "29/01/2022",
                "ICPCnotRequired": "false",
            },
            {  # 1 ignore DateOfInitialCPC notna
                "LAchildID": "child2",
                "DateOfInitialCPC": "26/05/2000",
                "S47ActualStartDate": "26/05/2001",
                "ICPCnotRequired": "false",
            },
            {  # 2 pass. more than 15 working days before ref date
                "LAchildID": "child3",
                "DateOfInitialCPC": pd.NA,
                "S47ActualStartDate": "25/03/2022",
                "ICPCnotRequired": "false",
            },
            {  # ignore S47ActualStartDate isna
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2000",
                "S47ActualStartDate": pd.NA,
                "ICPCnotRequired": "false",
            },
            {  # ignore
                "LAchildID": "child5",
                "DateOfInitialCPC": pd.NA,
                "S47ActualStartDate": pd.NA,
                "ICPCnotRequired": "true",
            },
            {  # 5 fail, no ICPCnotrequied as true or InitialCPC, and date is more than 15 days before end of census year
                "LAchildID": "child6",
                "DateOfInitialCPC": pd.NA,
                "S47ActualStartDate": "29/01/2022",
                "ICPCnotRequired": "0",
            },
        ]
    )
    # if rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    section47["DateOfInitialCPC"] = pd.to_datetime(
        section47["DateOfInitialCPC"], format="%d/%m/%Y", errors="coerce"
    )
    section47["S47ActualStartDate"] = pd.to_datetime(
        section47["S47ActualStartDate"], format="%d/%m/%Y", errors="coerce"
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {Section47: section47, Header: fake_header})

    # Use .type1_issues to check for the result of .push_type1_issues() which you used above.
    issues = result.type1_issues

    # get table name and check it. Replace ChildProtectionPlans with the name of your table.
    issue_table = issues.table
    assert issue_table == Section47

    # check that the right columns were returned. Replace CPPstartDate and CPPendDate with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [DateOfInitialCPC, S47ActualStartDate, ICPCnotReqiured]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 1 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 2
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    pd.to_datetime("29/01/2022", format="%d/%m/%Y", errors="coerce"),
                    "false",
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child6",
                    pd.to_datetime("29/01/2022", format="%d/%m/%Y", errors="coerce"),
                    "0",
                ),
                "ROW_ID": [5],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace 8675Q with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8675Q"
    assert (
        result.definition.message
        == "Please check and either amend data or provide a reason: S47 Enquiry started more than 15 working days before the end of the census year. However, there is no date of Initial Child Protection Conference."
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
UPN = ChildIdentifiers.UPN


@rule_definition(
    code="1550",
    module=CINTable.ChildIdentifiers,
    message="UPN invalid (character 13 not a recognised value)",
    affected_fields=[UPN],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildIdentifiers]

    # if <UPN> (N00001)) present Character 13 of <UPN> must be numeric or A-Z omitting I, O and S
    # Confirm length and value present
    df2 = df[(df["UPN"].str.len() == 13) & df["UPN"].notna()]

    # Valid characters
    valid = [
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "J",
        "K",
        "L",
        "M",
        "N",
        "P",
        "Q",
        "R",
        "T",
        "U",
        "V",
        "W",
        "Y",
        "X",
        "Z",
        "0",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
    ]

    failing_indices = df2[~df2["UPN"].str[12].isin(valid)].index

    rule_context.push_issue(table=ChildIdentifiers, field=UPN, row=failing_indices)


def test_validate():
    child_identifiers = pd.DataFrame(
        [["1234567891234"], ["123456789123I"], ["123456789123O"]], columns=[UPN]
    )

    result = run_rule(validate, {ChildIdentifiers: child_identifiers})

    issues = list(result.issues)
    assert len(issues) == 2
    assert issues == [
        IssueLocator(CINTable.ChildIdentifiers, UPN, 1),
        IssueLocator(CINTable.ChildIdentifiers, UPN, 2),
    ]

    assert result.definition.code == "1550"
    assert (
        result.definition.message == "UPN invalid (character 13 not a recognised value)"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

CINdetails = CINTable.CINdetails
ReasonForClosure = CINdetails.ReasonForClosure


# define characteristics of rule
@rule_definition(
    code="8640",
    module=CINTable.CINdetails,
    message="CIN Reason for closure code invalid (see Reason for Closure table in CIN Census code set)",
    affected_fields=[ReasonForClosure],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[CINdetails]

    # implement rule logic as described by the Github issue. Put the description as a comment above the implementation as shown.

    # Reason for Closure must be a valid reason for closure code value as shown in the list below

    valid_reason = ["RC1", "RC2", "RC3", "RC4", "RC5", "RC6", "RC7", "RC8", "RC9"]

    # Check if the Reason For Closure is not in the list of valid reasons and a value has been entered.

    df = df[
        ~(df["ReasonForClosure"].isin(valid_reason)) & df["ReasonForClosure"].notna()
    ]

    failing_indices = df.index

    # Replace CINdetails and ReasonForClosure with the table and column name concerned in your rule, respectively.
    # If there are multiple columns or table, make this sentence multiple times.
    rule_context.push_issue(
        table=CINdetails, field=ReasonForClosure, row=failing_indices
    )


def test_validate():
    # Sample test all will pass the validation with the exception RC13 and RC0 which are not valid Closure codes.
    fake_data = ["RC1", "RC13", pd.NA, "RC0"]

    fake_dataframe = pd.DataFrame({"ReasonForClosure": fake_data})

    # Run rule function passing in our sample test data
    result = run_rule(validate, {CINdetails: fake_dataframe})

    # The result contains a list of issues encountered
    issues = list(result.issues)
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issues) == 2
    # replace the table and column name as done earlier.
    # The last numbers represent the index values where you expect the sample data to fail the validation check.
    assert issues == [
        IssueLocator(CINTable.CINdetails, ReasonForClosure, 1),
        IssueLocator(CINTable.CINdetails, ReasonForClosure, 3),
    ]

    # Check that the rule definition is what you wrote in the context above.

    assert result.definition.code == "8640"
    assert (
        result.definition.message
        == "CIN Reason for closure code invalid (see Reason for Closure table in CIN Census code set)"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

Assessments = CINTable.Assessments
AssessmentAuthorisationDate = Assessments.AssessmentAuthorisationDate
AssessmentFactors = Assessments.AssessmentFactors
LAchildID = Assessments.LAchildID


@rule_definition(
    code="8614",
    module=CINTable.Assessments,
    message="Parental or child factors at assessment should only be present for a completed assessment.",
    affected_fields=[AssessmentAuthorisationDate, AssessmentFactors],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[Assessments]
    df.index.name = "ROW_ID"

    # If <AssessmentAuthorisationDate> (N00160) is missing,
    # <AssessmentFactors> (N00181) within the same module must also be missing
    # fails if AssessmentAuthoriationDate is null and AssessmentFactors  is not null
    condition = df[AssessmentAuthorisationDate].isna() & df[AssessmentFactors].notna()

    df_issues = df[condition].reset_index()

    link_id = tuple(
        zip(
            df_issues[LAchildID],
            df_issues[AssessmentAuthorisationDate],
            df_issues[AssessmentFactors],
        )
    )
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_1(
        table=Assessments,
        columns=[AssessmentAuthorisationDate, AssessmentFactors],
        row_df=df_issues,
    )


def test_validate():
    fake_data = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "AssessmentAuthorisationDate": "26/05/2000",
                "AssessmentFactors": "26/05/2000",
            },
            {
                "LAchildID": "child2",
                "AssessmentAuthorisationDate": "26/05/2000",
                "AssessmentFactors": ["6a", "6b"],
            },
            {
                "LAchildID": "child3",
                "AssessmentAuthorisationDate": pd.NA,
                "AssessmentFactors": "26/05/2000",
            },  # error - Authorisation date is null and Assessment factors is not
            {
                "LAchildID": "child3",
                "AssessmentAuthorisationDate": "26/05/2000",
                "AssessmentFactors": pd.NA,
            },
            {
                "LAchildID": "child4",
                "AssessmentAuthorisationDate": pd.NA,
                "AssessmentFactors": "25/05/2000",
            },  # error Authorisation date is null and assessment factors is not
            {
                "LAchildID": "child5",
                "AssessmentAuthorisationDate": pd.NA,
                "AssessmentFactors": pd.NA,
            },
        ]
    )

    result = run_rule(validate, {Assessments: fake_data})

    issues = result.type1_issues

    issue_table = issues.table
    assert issue_table == Assessments

    issue_columns = issues.columns
    assert issue_columns == [AssessmentAuthorisationDate, AssessmentFactors]

    issue_rows = issues.row_df
    assert len(issue_rows) == 2
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child3",
                    pd.NA,
                    "26/05/2000",
                ),
                "ROW_ID": [2],
            },
            {
                "ERROR_ID": (
                    "child4",
                    pd.NA,
                    "25/05/2000",
                ),
                "ROW_ID": [4],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "8614"
    assert (
        result.definition.message
        == "Parental or child factors at assessment should only be present for a completed assessment."
    )



"""
Rule number: '1105'
Module: Child protection plans
Rule details: Where present, the <CPPStartDate> (N00105) must be on or after the <CINReferralDate> (N00100)
Rule message: The child protection plan start date cannot be before the referral date

"""
from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

ChildProtectionPlans = CINTable.ChildProtectionPlans
CPPstartDate = ChildProtectionPlans.CPPstartDate
CPP_LAID = ChildProtectionPlans.LAchildID
CPP_CINdetailsID = ChildProtectionPlans.CINdetailsID

CINDetails = CINTable.CINdetails
CINreferralDate = CINDetails.CINreferralDate
CIN_LAID = CINDetails.LAchildID
CIN_CINdetailsID = CINDetails.CINdetailsID


@rule_definition(
    code="1105",
    module=CINTable.ChildProtectionPlans,
    message="The child protection plan start date cannot be before the referral date",
    affected_fields=[
        CPPstartDate,
        CINreferralDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_CPP = data_container[ChildProtectionPlans].copy()
    df_CIN = data_container[CINDetails].copy()

    df_CPP.index.name = "ROW_ID"
    df_CIN.index.name = "ROW_ID"

    df_CPP.reset_index(inplace=True)
    df_CIN.reset_index(inplace=True)

    # <CPPStartDate> (N00105) must be on or after the <CINReferralDate> (N00100)

    # Remove rows without CPP start date
    df_CPP = df_CPP[df_CPP[CPPstartDate].notna()]

    df = df_CPP.merge(
        df_CIN,
        left_on=["LAchildID", "CINdetailsID"],
        right_on=["LAchildID", "CINdetailsID"],
        how="left",
        suffixes=("_CPP", "_CIN"),
    )

    # Return those where dates don't align
    df = df[df["CINreferralDate"] > df["CPPstartDate"]].reset_index()

    df["ERROR_ID"] = tuple(zip(df["LAchildID"], df[CPPstartDate], df[CINreferralDate]))

    df_CPP_issues = (
        df_CPP.merge(df, left_on="ROW_ID", right_on="ROW_ID_CPP")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    df_CIN_issues = (
        df_CIN.merge(df, left_on="ROW_ID", right_on="ROW_ID_CIN")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=ChildProtectionPlans, columns=[CPPstartDate], row_df=df_CPP_issues
    )
    rule_context.push_type_2(
        table=CINDetails, columns=[CINreferralDate], row_df=df_CIN_issues
    )


def test_validate():
    sample_CPP = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CPPstartDate": "26/05/2000",  # Pass, same date
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child1",
                "CPPstartDate": "27/06/2002",  # Pass, after referall
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child3",
                "CPPstartDate": "07/02/1999",  # Fail, prior to referall
                "CINdetailsID": "cinID6",
            },
            {
                "LAchildID": "child2",
                "CPPstartDate": "26/05/2000",  # Fail, prior to referral
                "CINdetailsID": "cinID3",
            },
            {
                "LAchildID": "child3",
                "CPPstartDate": "26/05/2001",  # Pass, after referall
                "CINdetailsID": "cinID4",
            },
        ]
    )

    sample_CIN = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # Pass
                "CINreferralDate": "26/05/2000",
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child1",  # Pass
                "CINreferralDate": "26/05/2000",
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child3",  # Fail
                "CINreferralDate": "26/05/2000",
                "CINdetailsID": "cinID6",
            },
            {
                "LAchildID": "child2",  # Fail
                "CINreferralDate": "30/05/2000",
                "CINdetailsID": "cinID3",
            },
            {
                "LAchildID": "child3",  # Pass
                "CINreferralDate": "27/05/2000",
                "CINdetailsID": "cinID4",
            },
        ]
    )

    sample_CPP[CPPstartDate] = pd.to_datetime(
        sample_CPP[CPPstartDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_CIN[CINreferralDate] = pd.to_datetime(
        sample_CIN[CINreferralDate], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(
        validate,
        {
            ChildProtectionPlans: sample_CPP,
            CINDetails: sample_CIN,
        },
    )

    issues_list = result.type2_issues
    assert len(issues_list) == 2

    issues = issues_list[1]

    issue_table = issues.table
    assert issue_table == CINDetails

    issue_columns = issues.columns
    assert issue_columns == [CINreferralDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 2
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child2",  # ChildID
                    # Start date
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                    # Referral date
                    pd.to_datetime("30/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [3],
            },
            {
                "ERROR_ID": (
                    "child3",  # ChildID
                    # Start Date
                    pd.to_datetime("07/02/1999", format="%d/%m/%Y", errors="coerce"),
                    # Referral date
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [2],
            },
        ]
    )

    assert issue_rows.equals(expected_df)

    assert result.definition.code == "1105"
    assert (
        result.definition.message
        == "The child protection plan start date cannot be before the referral date"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

Assessments = CINTable.Assessments
AssessmentAuthorisationDate = Assessments.AssessmentAuthorisationDate
AssessmentActualStartDate = Assessments.AssessmentActualStartDate
CINdetailsID = Assessments.CINdetailsID
LAchildID = Assessments.LAchildID

Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


# define characteristics of rule
@rule_definition(
    code="8736",
    module=CINTable.Assessments,
    message="For an Assessment that has not been completed, the start date must fall within the census year",
    affected_fields=[AssessmentAuthorisationDate, AssessmentActualStartDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    df = data_container[Assessments]
    header = data_container[Header]
    ref_date_series = header[ReferenceDate]
    collection_start, collection_end = make_census_period(ref_date_series)

    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df.index.name = "ROW_ID"

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # Where present, if an <Assessments> group does not contain the <AssessmentAuthorisationDate> (N00160) then the <AssessmentActualStartDate> (N00159) must be on or between [Start_Of_Census_Year] and <ReferenceDate> (N00603)
    condition_1 = (df[CINdetailsID].notna()) & (df[AssessmentAuthorisationDate].isna())
    condition_2 = (collection_start <= df[AssessmentActualStartDate]) & (
        df[AssessmentActualStartDate] <= collection_end
    )

    # get all the data that fits the failing condition. Reset the index so that ROW_ID now becomes a column of df
    df_issues = df[condition_1 & ~condition_2].reset_index()

    # SUBMIT ERRORS
    # Generate a unique ID for each instance of an error.
    link_id = tuple(
        zip(
            df_issues[LAchildID],
            df_issues[CINdetailsID],
            df_issues[AssessmentActualStartDate],
        )
    )
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    rule_context.push_type_1(
        table=Assessments,
        columns=[AssessmentAuthorisationDate, AssessmentActualStartDate],
        row_df=df_issues,
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.

    fake_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2022"}]  # the census start date here will be 01/04/2021
    )

    child_assessments = pd.DataFrame(
        [
            {
                "CINdetailsID": pd.NA,
                "LAchildID": "child1",
                "AssessmentAuthorisationDate": pd.NA,
                "AssessmentActualStartDate": "27/05/2021",
            },
            {
                "CINdetailsID": "1",
                "LAchildID": "child2",
                "AssessmentAuthorisationDate": pd.NA,
                "AssessmentActualStartDate": "26/03/2020",
                # 1 error: start date is outside the reporting period
            },
            {
                "CINdetailsID": "1",
                "LAchildID": "child3",
                "AssessmentAuthorisationDate": "27/06/2021",
                "AssessmentActualStartDate": "27/05/2021",
            },
            {
                "CINdetailsID": "1",
                "LAchildID": "child4",
                "AssessmentAuthorisationDate": pd.NA,
                "AssessmentActualStartDate": "10/04/2024",
                # 3 error: start date is outside the reporting period
            },
            {
                "CINdetailsID": "1",
                "LAchildID": "child5",
                "AssessmentAuthorisationDate": pd.NA,
                "AssessmentActualStartDate": "27/05/2022",
                # 4 error: start date is outside the reporting period
            },
            {
                "CINdetailsID": pd.NA,
                "LAchildID": "child6",
                "AssessmentAuthorisationDate": pd.NA,
                "AssessmentActualStartDate": pd.NA,
            },
        ]
    )
    # if rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    child_assessments[AssessmentAuthorisationDate] = pd.to_datetime(
        child_assessments[AssessmentAuthorisationDate],
        format="%d/%m/%Y",
        errors="coerce",
    )
    child_assessments[AssessmentActualStartDate] = pd.to_datetime(
        child_assessments[AssessmentActualStartDate], format="%d/%m/%Y", errors="coerce"
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {Assessments: child_assessments, Header: fake_header})

    # Use .type1_issues to check for the result of .push_type1_issues() which you used above.
    issues = result.type1_issues

    issue_table = issues.table
    assert issue_table == Assessments

    # check that the right columns were returned. Replace CPPstartDate and CPPendDate with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [AssessmentAuthorisationDate, AssessmentActualStartDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 3 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 3
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child2",
                    "1",
                    pd.to_datetime("26/03/2020", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child4",
                    "1",
                    pd.to_datetime("10/04/2024", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [3],
            },
            {
                "ERROR_ID": (
                    "child5",
                    "1",
                    pd.to_datetime("27/05/2022", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [4],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace '8736' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8736"
    assert (
        result.definition.message
        == "For an Assessment that has not been completed, the start date must fall within the census year"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

CINdetails = CINTable.CINdetails
Section47 = CINTable.Section47

CINclosureDate = CINdetails.CINclosureDate
LAchildID = CINdetails.LAchildID
CINdetailsID = CINdetails.CINdetailsID
DateOfInitialCPC = Section47.DateOfInitialCPC
ICPCnotRequired = Section47.ICPCnotRequired


# define characteristics of rule
@rule_definition(
    code="8868",
    # replace CINdetails with the value in the module column of the excel sheet corresponding to this rule .
    # Note that even if multiple tables are involved, one table will be named in the module column.
    module=CINTable.CINdetails,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="CIN episode is shown as closed, however Section 47 enquiry is not shown as completed by ICPC date or ICPC not required flag",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[
        ICPCnotRequired,
        CINclosureDate,
        DateOfInitialCPC,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    df_47 = data_container[Section47].copy()
    df_cin = data_container[CINdetails].copy()

    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df_47.index.name = "ROW_ID"
    df_cin.index.name = "ROW_ID"

    df_47.reset_index(inplace=True)
    df_cin.reset_index(inplace=True)

    # If <CINclosureDate> (N00102) is present then all instances of the <Section47> group must either
    # i) include the <DateOfInitialCPC> (N00110) or
    # ii) include <ICPCnotRequired> (N00111) with a value of true or 1

    CINclosure_present = df_cin[df_cin[CINclosureDate].notna()]

    # Children with no s47 module must be excluded to prevent falsely flagging them, done usig inner merge.
    merged_df = CINclosure_present.merge(
        df_47,
        on=[
            LAchildID,
            "CINdetailsID",
        ],
        how="inner",
        suffixes=["_cin", "_47"],
    )

    # Checks DateOfInitialCPC from s47 model
    condition_1 = (merged_df["DateOfInitialCPC_47"].isna()) & (
        ~merged_df[ICPCnotRequired].isin(["1", "true"])
    )
    condition_2 = (merged_df["DateOfInitialCPC_47"].notna()) & (
        merged_df[ICPCnotRequired].isin(["1", "true"])
    )

    # get all the data that fits the failing condition.
    merged_df = merged_df[condition_1 | condition_2].reset_index()

    # create an identifier for each error instance.
    merged_df["ERROR_ID"] = tuple(zip(merged_df[LAchildID], merged_df[CINdetailsID]))

    # we can now map the suffixes columns to their corresponding source tables such that the failing ROW_IDs and ERROR_IDs exist per table.
    df_47_issues = (
        df_47.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_47")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_cin_issues = (
        df_cin.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # Ensure that you maintain the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.

    rule_context.push_type_2(
        table=Section47,
        columns=[DateOfInitialCPC, ICPCnotRequired],
        row_df=df_47_issues,
    )
    rule_context.push_type_2(
        table=CINdetails, columns=[CINclosureDate], row_df=df_cin_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_section47 = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "DateOfInitialCPC": pd.NA,  # 0 fail: DateOfInitialCPC missing and ICPCnotRequired false
                "ICPCnotRequired": "false",
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child1",
                "DateOfInitialCPC": pd.NA,  # 1 pass ICPCnotRequired true
                "ICPCnotRequired": "true",
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child2",
                "DateOfInitialCPC": pd.NA,
                "ICPCnotRequired": "1",  # 2 pass ICPCnotRequired true
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": "27/05/2000",  # 3 pass DateOfInitialCPC present
                "ICPCnotRequired": "false",
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/10/1999",  # 4 fail
                "ICPCnotRequired": "nottrue",
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2000",  # 5 pass DateOfInitialCPC present
                "ICPCnotRequired": "1",
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child4",
                "DateOfInitialCPC": pd.NA,
                "ICPCnotRequired": "false",  # 6 ignore
                "CINdetailsID": "cinID4",
            },
        ]
    )
    sample_cin = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # 0 fail: contains a section47 group that fails.
                "CINclosureDate": "26/10/1999",
                "CINdetailsID": "cinID1",
                "DateOfInitialCPC": "26/10/2001",
            },
            {
                "LAchildID": "child1",  # 1 pass
                "CINclosureDate": "26/05/2000",
                "CINdetailsID": "cinID2",
                "DateOfInitialCPC": pd.NA,
            },
            {
                "LAchildID": "child2",  # 2
                "CINclosureDate": "26/05/2000",
                "CINdetailsID": "cinID1",
                "DateOfInitialCPC": pd.NA,
            },
            {
                "LAchildID": "child3",  # 3 pass
                "CINclosureDate": "28/05/2000",
                "CINdetailsID": "cinID1",
                "DateOfInitialCPC": pd.NA,
            },
            {
                "LAchildID": "child3",  # 4 fail: contains a section47 group that fails.
                "CINclosureDate": "26/05/2000",
                "CINdetailsID": "cinID2",
                "DateOfInitialCPC": pd.NA,
            },
            {
                "LAchildID": "child3",  # 5 Pass, not present in section47 table so none of the values meets the requirements
                "CINclosureDate": "26/05/2003",
                "CINdetailsID": "cinID3",
                "DateOfInitialCPC": pd.NA,
            },
            {
                "LAchildID": "child4",
                "CINclosureDate": pd.NA,  # 6 ignore: date absent
                "CINdetailsID": "cinID4",
                "DateOfInitialCPC": pd.NA,
            },
        ]
    )
    # if rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    sample_cin[CINclosureDate] = pd.to_datetime(
        sample_cin[CINclosureDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_section47["DateOfInitialCPC"] = pd.to_datetime(
        sample_section47["DateOfInitialCPC"], format="%d/%m/%Y", errors="coerce"
    )
    sample_cin["DateOfInitialCPC"] = pd.to_datetime(
        sample_cin["DateOfInitialCPC"], format="%d/%m/%Y", errors="coerce"
    )

    # Run the rule function, passing in our sample data.
    result = run_rule(
        validate,
        {
            Section47: sample_section47,
            CINdetails: sample_cin,
        },
    )

    # Use .type2_issues to check for the result of .push_type2_issues() which you used above.
    issues_list = result.type2_issues
    assert len(issues_list) == 2
    # the function returns a list on NamedTuples where each NamedTuple contains (table, column_list, df_issues)
    # pick any table and check it's values. the tuple in location 1 will contain the CINdetails columns because that's the second thing pushed above.
    issues = issues_list[1]

    # get table name and check it. Replace CINdetails with the name of your table.
    issue_table = issues.table
    assert issue_table == CINdetails

    # check that the right columns were returned. Replace CINclosureDate with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [CINclosureDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 3 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 2
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on, in your zip, earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",  # ChildID
                    "cinID1",  # CINdetailsID
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child3",  # ChildID
                    "cinID2",  # CINdetailsID
                ),
                "ROW_ID": [4],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace '8868' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8868"
    assert (
        result.definition.message
        == "CIN episode is shown as closed, however Section 47 enquiry is not shown as completed by ICPC date or ICPC not required flag"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

ChildProtectionPlans = CINTable.ChildProtectionPlans
LAchildID_cpp = ChildProtectionPlans.LAchildID
CINdetailsID_cpp = ChildProtectionPlans.CINdetailsID

CINdetails = CINTable.CINdetails
ReferralNFA = CINdetails.ReferralNFA
LAchildID_details = CINdetails.LAchildID
CINdetailsID_details = CINdetails.CINdetailsID


# define characteristics of rule
@rule_definition(
    # write the rule code here
    code="8832",
    # replace ChildProtectionPlans with the value in the module column of the excel sheet corresponding to this rule .
    # Note that even if multiple tables are involved, one table will be named in the module column.
    module=CINTable.ChildProtectionPlans,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="Child Protection details provided for a referral with no further action.",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[
        CINdetailsID_cpp,
        ReferralNFA,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    # Replace ChildProtectionPlans with the name of the table you need.
    df_cpp = data_container[ChildProtectionPlans]
    df_CINdetails = data_container[CINdetails]

    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df_cpp.index.name = "ROW_ID"
    df_CINdetails.index.name = "ROW_ID"

    # Resetting the index causes the ROW_IDs to become columns of their respective DataFrames
    # so that they can come along when the merge is done.
    df_cpp.reset_index(inplace=True)
    df_CINdetails.reset_index(inplace=True)

    # If a <CINdetails> module has <ReferralNFA> (N00112) = true or 1, then there should be no Child Protection module present

    # Excluding rows with false or 0 <ReferralNFA> to fix bug where they were flagged as failing
    # TODO remove this step when proven to be redundant.
    df_CINdetails = df_CINdetails[
        ~(df_CINdetails[ReferralNFA].str.lower() == "false")
        | ~(df_CINdetails[ReferralNFA].astype(str) == "0")
    ]

    df_CINdetails = df_CINdetails[
        (df_CINdetails[ReferralNFA].str.lower() == "true")
        | (df_CINdetails[ReferralNFA].astype(str) == "1")
    ]

    #  Merge tables to get corresponding CP plan group and reviews
    df_merged = df_CINdetails.merge(
        df_cpp,
        on=["LAchildID", "CINdetailsID"],
        how="inner",
        suffixes=("_cpp", "_cin"),
    )
    # any rows found in df_merged are ones where <ReferralNFA> (N00112) = true or 1 and yet the child existed in the CINdetails table.
    df_merged = df_merged.reset_index()

    # create an identifier for each error instance.
    # In this case, the rule is checked for each CPPstartDate, in each CPplanDates group (differentiated by CP dates), in each child (differentiated by LAchildID)
    # So, a combination of LAchildID, CPPstartDate and CPPreviewDate identifies and error instance.
    df_merged["ERROR_ID"] = tuple(
        zip(
            df_merged["LAchildID"],
            df_merged["CINdetailsID"],
        )
    )

    # The merges were done on copies of cpp_df and reviews_df so that the column names in dataframes themselves aren't affected by the suffixes.
    # we can now map the suffixes columns to their corresponding source tables such that the failing ROW_IDs and ERROR_IDs exist per table.
    df_cpp_issues = (
        df_cpp.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cpp")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_cin_issues = (
        df_CINdetails.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # Ensure that you maintain the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_2(
        table=ChildProtectionPlans, columns=[LAchildID_cpp], row_df=df_cpp_issues
    )
    rule_context.push_type_2(
        table=CINdetails, columns=[ReferralNFA], row_df=df_cin_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_cpp = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINdetailsID": "CDID1",
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "CDID2",
            },
            # {
            #     "LAchildID": "child3",
            #     "CINdetailsID": "CDID6",
            # },
            {
                "LAchildID": "child4",  # ignored
                "CINdetailsID": "CDID0",
            },
            {
                "LAchildID": "child5",  # ignored, ReferralNFA is neither "1" nor "true"
                "CINdetailsID": "CDID0",
            },
        ]
    )
    sample_cin = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # Fail, module present
                "CINdetailsID": "CDID1",
                "ReferralNFA": "1",
            },
            {
                "LAchildID": "child1",  # Pass, no referralNFA
                "CINdetailsID": "CDID2",
                "ReferralNFA": pd.NA,
            },
            {
                "LAchildID": "child3",  # Pass, no module
                "CINdetailsID": "CDID6",
                "ReferralNFA": "1",
            },
            {
                "LAchildID": "child4",  # ignored, ReferralNFA is neither "1" nor "true"
                "CINdetailsID": "CDID0",
                "ReferralNFA": "false",
            },
            {
                "LAchildID": "child5",  # ignored, ReferralNFA is neither "1" nor "true"
                "CINdetailsID": "CDID0",
                "ReferralNFA": 0,
            },
        ]
    )

    # Run the rule function, passing in our sample data.
    result = run_rule(
        validate,
        {
            ChildProtectionPlans: sample_cpp,
            CINdetails: sample_cin,
        },
    )

    # Use .type2_issues to check for the result of .push_type2_issues() which you used above.
    issues_list = result.type2_issues
    assert len(issues_list) == 2
    # the function returns a list on NamedTuples where each NamedTuple contains (table, column_list, df_issues)
    # pick any table and check it's values. the tuple in location 1 will contain the CINdetails columns because that's the second thing pushed above.
    issues = issues_list[1]

    # get table name and check it. Replace Reviews with the name of your table.
    issue_table = issues.table
    assert issue_table == CINdetails

    issue_columns = issues.columns
    assert issue_columns == [ReferralNFA]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 1 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 1
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on, in your zip, earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",  # ChildID
                    "CDID1",
                ),
                "ROW_ID": [0],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace '8832' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8832"
    assert (
        result.definition.message
        == "Child Protection details provided for a referral with no further action."
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

ChildProtectionPlans = CINTable.ChildProtectionPlans
NumberOfPreviousCPP = ChildProtectionPlans.NumberOfPreviousCPP


# define characteristics of rule
@rule_definition(
    code="8730",
    module=CINTable.ChildProtectionPlans,
    message="Total Number of previous Child Protection Plans missing",
    affected_fields=[NumberOfPreviousCPP],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # Replace ChildProtectionPlans with the name of the table you need.
    df = data_container[ChildProtectionPlans]

    # implement rule logic as described by the Github issue. Put the description as a comment above the implementation as shown.

    # Where a CPP module is present, <NumberOfPreviousCPP> (N00106) must be greater than or equal to zero
    # Change the line below to ensure values are >=0 ie not null

    failing_indices = df[
        (df[NumberOfPreviousCPP].isna()) | (df[NumberOfPreviousCPP].astype("Int64") < 0)
    ].index
    # Int64 dtype is used instead of int because it tolerates the possibility of NaN values in the column
    # nullable integers ==  Int64 dtype

    rule_context.push_issue(
        table=ChildProtectionPlans, field=NumberOfPreviousCPP, row=failing_indices
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    fake_data = pd.DataFrame(
        [[1234], [pd.NA], [pd.NA], [-1]], columns=[NumberOfPreviousCPP]
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {ChildProtectionPlans: fake_data})
    # The result contains a list of issues encountered
    issues = list(result.issues)
    # replace 3 with the number of failing points you expect from the sample data.
    assert len(issues) == 3
    # replace the table and column name as done earlier.
    # The last numbers represent the index values where you expect the sample data to fail the validation check.
    assert issues == [
        IssueLocator(CINTable.ChildProtectionPlans, NumberOfPreviousCPP, 1),
        IssueLocator(CINTable.ChildProtectionPlans, NumberOfPreviousCPP, 2),
        IssueLocator(CINTable.ChildProtectionPlans, NumberOfPreviousCPP, 3),
    ]

    # Check that the rule definition is what you wrote in the context above.

    # replace '8730' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8730"
    assert (
        result.definition.message
        == "Total Number of previous Child Protection Plans missing"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

Section47 = CINTable.Section47
LAchildID = Section47.LAchildID
S47ActualStartDate = Section47.S47ActualStartDate
DateOfInitialCPC = Section47.DateOfInitialCPC
ICPCnotRequired = Section47.ICPCnotRequired

Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


# define characteristics of rule
@rule_definition(
    code="8740",
    module=CINTable.Section47,
    message="For a Section 47 Enquiry that has not held the Initial Child Protection Conference by the end of the census year, the start date must fall within the census year",
    affected_fields=[S47ActualStartDate, DateOfInitialCPC, ICPCnotRequired],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    df = data_container[Section47]
    # Rename the index so that the initial row positions can be kept intact.
    df.index.name = "ROW_ID"

    # ReferenceDate exists in the heder table so we get header table too.
    df_ref = data_container[Header]
    ref_date_series = df_ref[ReferenceDate]

    # the make_census_period function generates the start and end date so that you don't have to do it each time.
    collection_start, reference_date = make_census_period(ref_date_series)

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # If a <Section47> group does not contain the <DateOfInitialCPC> (N00110) and <ICPCnotRequired> (N00111) is false
    # then the <S47ActualStartDate> (N00148) must be on or between [Start_Of_Census_Year] and <ReferenceDate> (N00603)
    condition = (
        df[DateOfInitialCPC].isna()
        & (df[ICPCnotRequired].astype(str) == "0")
        & (
            (df[S47ActualStartDate] < collection_start)
            | (df[S47ActualStartDate] > reference_date)
        )
    )
    # get all the data that fits the failing condition. Reset the index so that ROW_ID now becomes a column of df
    df_issues = df[condition].reset_index()

    # SUBMIT ERRORS
    # Generate a unique ID for each instance of an error.
    link_id = tuple(zip(df_issues[LAchildID], df_issues[S47ActualStartDate]))
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys="False")["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    # Ensure that you do not change the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_1(
        table=Section47,
        columns=[S47ActualStartDate, DateOfInitialCPC, ICPCnotRequired],
        row_df=df_issues,
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.

    fake_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2001"}]  # the census start date here will be 01/04/2000
    )

    section_47s = pd.DataFrame(
        [
            {  # 0 - Ignore
                "LAchildID": "child1",
                "S47ActualStartDate": "26/05/2000",
                "DateOfInitialCPC": "26/05/2000",
                "ICPCnotRequired": "0",
            },
            {  # 1 - Ignore
                "LAchildID": "child1",
                "S47ActualStartDate": "26/05/2000",
                "DateOfInitialCPC": "26/05/2000",
                "ICPCnotRequired": "1",
            },
            {  # 2 - Ignore
                "LAchildID": "child1",
                "S47ActualStartDate": "26/05/2000",
                "DateOfInitialCPC": pd.NA,
                "ICPCnotRequired": "1",
            },
            {  # 3 - Pass
                "LAchildID": "child1",
                "S47ActualStartDate": "26/05/2000",
                "DateOfInitialCPC": pd.NA,
                "ICPCnotRequired": "0",
            },
            {  # 4 - Fail
                "LAchildID": "child1",
                "S47ActualStartDate": "26/05/1999",
                "DateOfInitialCPC": pd.NA,
                "ICPCnotRequired": "0",
            },
            {  # 5 - Fail
                "LAchildID": "child1",
                "S47ActualStartDate": "26/05/1999",
                "DateOfInitialCPC": pd.NA,
                "ICPCnotRequired": "0",
            },
            {  # 6 - Ignore
                "LAchildID": "child1",
                "S47ActualStartDate": "26/05/1999",
                "DateOfInitialCPC": pd.NA,
                "ICPCnotRequired": "1",
            },
        ]
    )
    # if rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    section_47s[S47ActualStartDate] = pd.to_datetime(
        section_47s[S47ActualStartDate], format="%d/%m/%Y", errors="coerce"
    )
    section_47s[DateOfInitialCPC] = pd.to_datetime(
        section_47s[DateOfInitialCPC], format="%d/%m/%Y", errors="coerce"
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {Section47: section_47s, Header: fake_header})

    # Use .type1_issues to check for the result of .push_type1_issues() which you used above.
    issues = result.type1_issues

    issue_table = issues.table
    assert issue_table == Section47

    issue_columns = issues.columns
    assert issue_columns == [S47ActualStartDate, DateOfInitialCPC, ICPCnotRequired]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 1 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 1
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    pd.to_datetime("26/05/1999", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [4, 5],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace '8740' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8740"
    assert (
        result.definition.message
        == "For a Section 47 Enquiry that has not held the Initial Child Protection Conference by the end of the census year, the start date must fall within the census year"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.test_engine import run_rule

Assessments = CINTable.Assessments
Section47 = CINTable.Section47


LAchildID = Assessments.LAchildID
CINdetailsID = Assessments.CINdetailsID


@rule_definition(
    code="2991Q",
    module=CINTable.CINdetails,
    rule_type=RuleType.QUERY,
    message="Please check and either amend data or provide a reason: A Section 47 module is recorded and there is no assessment on the episode",
    affected_fields=[
        CINdetailsID,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_ass = data_container[Assessments].copy()
    df_47 = data_container[Section47].copy()

    df_ass.index.name = "ROW_ID"
    df_47.index.name = "ROW_ID"

    df_ass.reset_index(inplace=True)
    df_47.reset_index(inplace=True)

    #  If <Section47> module is present then <Assessment> module should be present.
    merged_df = df_47.merge(
        df_ass,
        on=[LAchildID, CINdetailsID],
        suffixes=["_47", "_ass"],
        how="outer",
        indicator=True,
    )
    merged_df = merged_df[merged_df["_merge"] == "left_only"]

    merged_df["ERROR_ID"] = tuple(
        zip(
            merged_df[LAchildID],
            merged_df[CINdetailsID],
        )
    )
    df_ass_issues = (
        df_ass.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_ass")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    df_47_issues = (
        df_47.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_47")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=Assessments, columns=[LAchildID], row_df=df_ass_issues
    )
    rule_context.push_type_2(table=Section47, columns=[LAchildID], row_df=df_47_issues)


def test_validate():
    sample_ass = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CPPstartDate": "26/05/2000",
                "CINdetailsID": pd.NA,
            },
            {
                "LAchildID": "child1",
                "CPPstartDate": "27/06/2002",
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child3",
                "CPPstartDate": "26/05/2000",
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child3",
                "CPPstartDate": pd.NA,
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child3",
                "CPPstartDate": "07/02/2001",
                "CINdetailsID": "cinID3",
            },
            {
                "LAchildID": "child3",
                "CPPstartDate": "14/03/2001",
                "CINdetailsID": "cinID4",
            },
            {
                "LAchildID": "child4",
                "CPPstartDate": "14/03/2001",
                "CINdetailsID": "cinID4",
            },
        ]
    )
    sample_section47 = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # 0 fail. No assessment
                "DateOfInitialCPC": "26/05/2000",
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child1",
                "DateOfInitialCPC": "26/05/2000",
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child2",  # 2 fail. No assessment
                "DateOfInitialCPC": "30/05/2000",
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": "27/05/2000",
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2000",
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2000",
                "CINdetailsID": "cinID3",
            },
            {
                "LAchildID": "child3",  # 6 fail. No assessment
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": pd.NA,
            },
        ]
    )

    result = run_rule(
        validate,
        {
            Assessments: sample_ass,
            Section47: sample_section47,
        },
    )

    issues_list = result.type2_issues
    assert len(issues_list) == 2
    issues = issues_list[1]

    issue_table = issues.table
    assert issue_table == Section47

    issue_columns = issues.columns
    assert issue_columns == [LAchildID]

    issue_rows = issues.row_df
    assert len(issue_rows) == 3
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    "cinID1",
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child2",
                    "cinID1",
                ),
                "ROW_ID": [2],
            },
            {
                "ERROR_ID": (
                    "child3",
                    pd.NA,
                ),
                "ROW_ID": [6],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "2991Q"
    assert (
        result.definition.message
        == "Please check and either amend data or provide a reason: A Section 47 module is recorded and there is no assessment on the episode"
    )



from pathlib import Path

from cin_validator.rule_engine import RuleDefinition, YearConfig
from cin_validator.rules.cin2023_24 import registry as prev_registry
from cin_validator.rules.ruleset_utils import (
    extract_validator_functions,
    update_validator_functions,
)

files = Path(__file__).parent.glob("*.py")
this_year_validator_funcs: dict[str, RuleDefinition] = extract_validator_functions(
    files
)
# if any rules need to be deleted, add their codes as strings into del_list
del_list: list[str] = []
this_year_config = YearConfig(
    deleted=del_list, added_or_modified=this_year_validator_funcs
)

registry = update_validator_functions(prev_registry, this_year_config)
__all__ = ["registry"]



"""
Rule number: '8585Q'
Module: CIN plan dates
Rule details: If <ReasonForClosure> (N00103) = RC2 (Died) then a valid <PersonDeathDate> (N00108) must be present.
Rule message: Please check: CIN episode shows Died as the Closure Reason, however child has no recorded Date of Death

"""

from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.rules.cin2022_23.rule_8925 import LAchildID
from cin_validator.test_engine import run_rule

CINdetails = CINTable.CINdetails
LAchildID = CINdetails.LAchildID
ReasonForClosure = CINdetails.ReasonForClosure

ChildIdentifiers = CINTable.ChildIdentifiers
LAchildID = ChildIdentifiers.LAchildID
PersonDeathDate = ChildIdentifiers.PersonDeathDate


@rule_definition(
    code="8585Q",
    module=CINTable.CINdetails,
    rule_type=RuleType.QUERY,
    message="Please check and either amend or provide a reason: CIN episode shows Died as the Closure Reason, however child has no recorded Date of Death",
    affected_fields=[
        ReasonForClosure,
        PersonDeathDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_CI = data_container[ChildIdentifiers].copy()
    df_CIN = data_container[CINdetails].copy()

    df_CI.index.name = "ROW_ID"
    df_CIN.index.name = "ROW_ID"

    df_CI.reset_index(inplace=True)
    df_CIN.reset_index(inplace=True)

    # <ReasonForClosure> (N00103) = RC2 (Died) then a valid <PersonDeathDate> (N00108) must be present.
    df = df_CI.merge(
        df_CIN,
        left_on=["LAchildID"],
        right_on=["LAchildID"],
        how="left",
        suffixes=("_CPP", "_CIN"),
    )

    df = df[df[ReasonForClosure] == "RC2"]
    df = df[df[PersonDeathDate].isna()].reset_index()

    df["ERROR_ID"] = tuple(
        zip(df[LAchildID], df[ReasonForClosure], df[PersonDeathDate])
    )

    df_CI_issues = (
        df_CI.merge(df, left_on="ROW_ID", right_on="ROW_ID_CPP")
        .groupby("ERROR_ID")["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    df_CIN_issues = (
        df_CIN.merge(df, left_on="ROW_ID", right_on="ROW_ID_CIN")
        .groupby("ERROR_ID")["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=ChildIdentifiers, columns=[PersonDeathDate], row_df=df_CI_issues
    )
    rule_context.push_type_2(
        table=CINdetails, columns=[ReasonForClosure], row_df=df_CIN_issues
    )


def test_validate():
    sample_CIN = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # Pass, closed due to death and death date
                "ReasonForClosure": "RC2",
            },
            {
                "LAchildID": "child2",  # Pass, ReasonForClosure not death
                "ReasonForClosure": "abc",
            },
            {
                "LAchildID": "child3",  # Fail, no death date
                "ReasonForClosure": "RC2",
            },
        ]
    )

    sample_CI = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "PersonDeathDate": "26/05/2000",  # Pass
            },
            {
                "LAchildID": "child2",
                "PersonDeathDate": "27/06/2002",  # Pass
            },
            {
                "LAchildID": "child3",
                "PersonDeathDate": pd.NA,  # Fail
            },
        ]
    )

    sample_CI[PersonDeathDate] = pd.to_datetime(
        sample_CI[PersonDeathDate], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(
        validate,
        {
            ChildIdentifiers: sample_CI,
            CINdetails: sample_CIN,
        },
    )

    issues_list = result.type2_issues
    assert len(issues_list) == 2

    issues = issues_list[1]

    issue_table = issues.table
    assert issue_table == CINdetails

    issue_columns = issues.columns
    assert issue_columns == [ReasonForClosure]

    issue_rows = issues.row_df
    assert len(issue_rows) == 1
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    # ChildID
                    "child3",
                    # Reason for closure
                    "RC2",
                    # Death date
                    pd.to_datetime(pd.NA, format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [2],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "8585Q"
    assert (
        result.definition.message
        == "Please check and either amend or provide a reason: CIN episode shows Died as the Closure Reason, however child has no recorded Date of Death"
    )



import copy
import xml.etree.ElementTree as ET
from typing import Optional

import pandas as pd

from cin_validator.ingress import XMLtoCSV
from cin_validator.rule_engine import CINTable, RuleContext, RuleDefinition
from cin_validator.utils import process_date_columns

pd.options.mode.chained_assignment = None
# Suppresses false-positive SettingWithCopyError when column types are changes in the include_issue_child function.
# https://stackoverflow.com/questions/20625582/how-to-deal-with-settingwithcopywarning-in-pandas


def enum_keys(dict_input: dict):
    """
    Convert keys of a dictionary to its corresponding CINTable format.
    :param dict dict_input: dictionary of dataframes of CIN data
    :return dict enumed_dict: same data content with keys replaced.
    """
    enumed_dict = {}
    for enum_key in CINTable:
        # if enum_key == CINTable.Header, then enum_key.name == Header
        enumed_dict[enum_key] = dict_input[str(enum_key.name)]
    return enumed_dict


def convert_data(root: ET.Element):
    """
    Takes input data and processes it for validation.

    This function takes input XML data, and uses ElementTree, and the custom class
    XMLtoCSV to process the data into tables for validation.

    :param XML root: root created by parsing the user's xml file.
    :returns: dict of DataFrames - each representing a CIN table.
    :rtype: Dictionary
    """

    # generate tables
    data_files = XMLtoCSV(root)

    # return tables
    cin_tables = {
        "Header": data_files.Header,
        "ChildIdentifiers": data_files.ChildIdentifiers,
        "ChildCharacteristics": data_files.ChildCharacteristics,
        "ChildProtectionPlans": data_files.ChildProtectionPlans,
        "CINdetails": data_files.CINdetails,
        "CINplanDates": data_files.CINplanDates,
        "Reviews": data_files.Reviews,
        "Section47": data_files.Section47,
        "Assessments": data_files.Assessments,
        "AssessmentFactorsList": data_files.AssessmentFactorsList,
        "Disabilities": data_files.Disabilities,
    }
    return cin_tables


def process_data(cin_tables: dict):
    """
    formats date columns
    :param dict cin_tables: data to be converted
    :return dict cin_tables_dict: original dataframes where date columns have been formatted.
    """

    # format all date columns in tables
    cin_tables_dict = {
        name: process_date_columns(table) for name, table in cin_tables.items()
    }

    return cin_tables_dict


def include_issue_child(issue_df: pd.DataFrame, cin_data: dict):
    """
    :param DataFrame issue_df: complete data about all issue locations.
    :param dict cin_data: dictionary of dataframes generated when cin xml is converted to tabular format.
    """

    try:
        la_level_issues = issue_df[issue_df["tables_affected"].isna()]
    except:
        # if no error locations were found, i.e issue_df doesn't exist, this allows an empty dataframe to be processed.
        return issue_df

    header_issues = issue_df[issue_df["tables_affected"] == "Header"]
    tables_with_childid = [la_level_issues, header_issues]
    for table in issue_df["tables_affected"].dropna().unique():
        if table == "Header":
            # the header table doesn't contain child id. It is like metadata
            continue
        table_df = issue_df[issue_df["tables_affected"] == table]

        # get index values of the rows that fail.
        # some ROW_ID values exist as ints and others as strs. Unify so that .unique() doesn't contain doubles.
        table_rows = table_df["ROW_ID"].astype("int").unique()

        # naming the index of the data allows it to be mapped back to the issue_df
        table_data = cin_data[table]
        table_data.index.name = "ROW_ID"
        table_data.reset_index(inplace=True)
        # select the data for the rows with appear in issue_df and get the child ids
        linker_df = table_data.iloc[table_rows][["LAchildID", "ROW_ID"]]

        # work around: ensure that columns from both sources have the same type to prevent merge error
        table_df["ROW_ID"] = table_df["ROW_ID"].astype("int64")
        linker_df["ROW_ID"] = linker_df["ROW_ID"].astype("int64")

        if not linker_df.empty:
            # if failing locations have been found, remove the dummy LAchildID column to make way for the real one.
            table_df.drop(columns="LAchildID", inplace=True)

        # map the child ids back to issue_df
        table_df = table_df.merge(linker_df, on=["ROW_ID"], how="left")

        # save the result
        tables_with_childid.append(table_df)

    # regenerate issue_df from its updated constituent tables
    issue_df = pd.concat(tables_with_childid)

    return issue_df


def create_user_report(issue_df: pd.DataFrame, cin_data: dict):
    """
    A good report should tell the user what failed, where it failed and why it failed.
    The report generated by this function contains table-column-value combinations to answer the former
    and rule code-description combinations to answer the latter.

    :param pd.DataFrame issue_df: in which child IDs have been added.
    :param dict cin_data: dataframes of user's input data.
    :return user_report: dataframe containing issue locations and specific values that fail in those locations.

    """
    try:
        no_table = issue_df[issue_df["tables_affected"].isna()]
    except:
        # in the case where issue_df is empty, return an empty user report.
        return pd.DataFrame()

    reports = []
    for table in issue_df["tables_affected"].dropna().unique():
        table_issues = issue_df[issue_df["tables_affected"] == table]

        table_reports = []
        for column in table_issues["columns_affected"].unique():
            only_column = table_issues[table_issues["columns_affected"] == column]
            column_rows = only_column["ROW_ID"].unique().astype("int")

            column_data = cin_data[table][column]
            # fancy indexing. get all the values for a sequence of row positions in a column.
            column_values = column_data[column_rows]
            column_values.rename("value_flagged", inplace=True)
            column_values.index.name = "ROW_ID"
            values_df = column_values.reset_index()
            values_df = values_df.assign(ROW_ID=values_df["ROW_ID"].astype("object"))

            # work around: ensure that columns from both sources (user data and rule output) have the same type to prevent merge error
            only_column["ROW_ID"] = only_column["ROW_ID"].astype("int64")
            values_df["ROW_ID"] = values_df["ROW_ID"].astype("int64")
            report_df = only_column.merge(values_df, on="ROW_ID")

            table_reports.append(report_df)

        reports.extend(table_reports)
    # add in the la-level locations
    reports.append(no_table)

    # ensure that all required column names will be present in the result.
    full_report_cols_df = pd.DataFrame(
        columns=[
            "ERROR_ID",
            "LAchildID",
            "rule_code",
            "tables_affected",
            "columns_affected",
            "ROW_ID",
            "value_flagged",
            "rule_description",
        ]
    )
    reports.append(full_report_cols_df)

    full_report = pd.concat(reports, ignore_index=True)

    # columns of interest are filtered and arranged in the desired order. values unified under str datatype.
    user_report = full_report[
        [
            "ERROR_ID",
            "LAchildID",
            "rule_code",
            "tables_affected",
            "columns_affected",
            "ROW_ID",
            "value_flagged",
            "rule_description",
        ]
    ]

    def datetime_to_str(element):
        if isinstance(element, pd.Timestamp):
            # convert datetime elements to str date values
            return str(element.strftime("%Y-%m-%d"))
        elif isinstance(element, tuple):
            # loop through tuples and convert each element accordingly. mostly in ERROR_ID column.
            return tuple(map(datetime_to_str, element))
        else:
            # ensure all other elements are strings too.
            return str(element)

    user_report = user_report.applymap(datetime_to_str)

    # Related issue locations should be displayed next to each other.
    user_report.sort_values(
        [
            "LAchildID",
            "ERROR_ID",
            "tables_affected",
            "columns_affected",
        ],
        inplace=True,
        ignore_index=True,
    )

    user_report.drop_duplicates(
        ["LAchildID", "rule_code", "columns_affected", "ROW_ID"], inplace=True
    )

    return user_report


class CinValidator:
    """
    A class to contain the process of CIN validation. Generates error reports as dataframes.

    :param any data_files: Data files for validation, either a DataContainerWrapper object, or a
        dictionary of DataFrames.
    :param dir ruleset: The directory containing the validation rules to be run according to the year in which they were published.
    """

    def __init__(
        self,
        data_files,
        ruleset_registry,
        selected_rules: Optional[list[str]] = None,
    ) -> None:
        """
        Initialises CinValidator class.

        Creates DataFrame containing error report, and allows selection of individual instances of error using ERROR_ID

        :param list ruleset: The list of rules used in an individual validation session.
            Refers to rules in particular subdirectories of the rules directory.
        :param any data_files: The data extracted from input XML (or CSV) for validation.
        :param str issue_id: Can be used to choose a particular instance of an error using ERROR_ID.
        :param list selected_rules: array of rule codes (as strings) selected by the user. Determines what rules should be run.
        :returns: DataFrame of error report which could be a filtered version if issue_id is input.
        :rtype: DataFrame
        """

        self.data_files = data_files
        self.ruleset_registry = ruleset_registry

        # save independent version of data to be used in report.
        raw_data = copy.deepcopy(self.data_files)

        # run
        self.create_issue_report_df(selected_rules)

        # add child_id to issue location report.
        self.full_issue_df: pd.DataFrame = include_issue_child(
            self.full_issue_df, raw_data
        )
        self.user_report = create_user_report(self.full_issue_df, raw_data)

        # regularise full_issue_df
        self.full_issue_df.rename(columns={"ROW_ID": "row_id"}, inplace=True)
        self.full_issue_df.rename(columns={"LAchildID": "child_id"}, inplace=True)
        self.full_issue_df.drop(columns=["ERROR_ID"], inplace=True, errors="ignore")
        self.full_issue_df.drop_duplicates(
            ["child_id", "rule_code", "columns_affected", "row_id"], inplace=True
        )
        self.full_issue_df.reset_index(drop=True, inplace=True)

        # remove header issues from main dataframe and transfer to no-child-id dataframe
        header_issues = self.full_issue_df[
            self.full_issue_df["tables_affected"] == "Header"
        ]
        self.full_issue_df = self.full_issue_df[
            self.full_issue_df["tables_affected"] != "Header"
        ]
        # combine multichild issues
        self.multichild_issues = pd.concat([header_issues, self.la_rule_issues])[
            ["rule_code", "rule_description"]
        ]

    def get_rules_to_run(
        self, registry, selected_rules: Optional[list[str]] = None
    ) -> list[RuleDefinition]:
        """
        Filters rules to be run based on user's selection in the frontend.
        :param Registry-class registry: record of all existing rules in rule pack
        :param list selected_rules: array of rule codes as strings
        """
        if selected_rules:
            rules_to_run = [
                rule for _, rule in registry.items() if str(rule.code) in selected_rules
            ]
            return rules_to_run
        else:
            return registry.values()

    def process_issues(self, rule: RuleDefinition, ctx: RuleContext):
        """
        process result of running a rule on the user's data.

        :param RuleDefinition-class rule: the rule that was run on the data
        :param RuleContext-object ctx: "manages state" per rule. contains updated issue_dfs if any were added when rule was run on the data.
        :returns : None

        """
        issue_dfs_per_rule = pd.Series(
            [
                ctx.type_zero_issues,
                ctx.type_one_issues,
                ctx.type_two_issues,
                ctx.type_three_issues,
                ctx.la_level_issues,
            ]
        )
        # error_df_lengths is a list of lengths of all elements in issue_dfs_per_rule respectively.
        error_df_lengths = pd.Series([len(x) for x in issue_dfs_per_rule])
        if error_df_lengths.max() == 0:
            # if the rule didn't push to any of the issue accumulators, then it didn't find any issues in the file.
            self.rules_passed.append(rule.code)
        elif error_df_lengths.idxmax() == 4:
            # If the maximum value is in position 4, this is a return level validation rule.
            # It has no locations attached so it is only displayed in the rule descriptions.
            self.la_rules_broken.append(issue_dfs_per_rule[4])
        else:
            # get the rule type based on which attribute had elements pushed to it (i.e non-zero length)
            # its corresponding error_df can be found by issue_dfs_per_rule[ind]
            ind = error_df_lengths.idxmax()

            issue_dict = {
                "code": rule.code,
                "number": error_df_lengths[ind],
                "type": ind,
            }
            issue_dict_df = pd.DataFrame([issue_dict])
            self.issue_instances = pd.concat(
                [self.issue_instances, issue_dict_df], ignore_index=True
            )

            # add the rule's code and description to it's error_df
            issue_dfs_per_rule[ind]["rule_code"] = rule.code
            issue_dfs_per_rule[ind]["rule_description"] = rule.message

            # temporary: add rule type to track if all types are in df.
            issue_dfs_per_rule[ind]["rule_type"] = ind

            # combine this rule's error_df with the cummulative error_df
            self.full_issue_df = pd.concat(
                [self.full_issue_df, issue_dfs_per_rule[ind]],
                ignore_index=True,
            )

            # Elements of the rule_descriptors df to explain error codes
            self.rules_broken.append(rule.code)
            self.rule_messages.append(f"{str(rule.code)} - {rule.message}")

    def create_issue_report_df(self, selected_rules: Optional[list[str]] = None):
        """
        Creates report of errors found when validating CIN data input to
        the tool.

        This function takes the errors/rule violations reported by individual validation rule functions,
        including table, field, and index locations of errors. It is important that it uses deepcopy
        on the data per rule as some rules alter original data when only a standard .copy() function
        is used. It runs through every rule in the registry and:

        >Creates lists of rules passed, broken, and relevant messages.
        >Returns a dataframe of issue instances for broken validation rules.
        >Returns a dictionary of all rules codes and relevant messages.

        :param DataFrame issue_instances: issues found in validation.
        :param DataFrame all_rules_issue_locs: issue locations
        :param list rules_broken: An empty list which is populated with the codes of the rules that trigger issues in the data during validation.
        :param list la_rules_broken: An empty list which is populated with the list of LA rules that fail validation.
        :param list rules_passed: An empty list of rules passed, populated with rules with no validation errors.
        :returns: DataFrame of instances and locations of validation rule violations from data input via FE or CLI.
        :rtype: DataFrame
        :raises: Errors with rules that raise errors when validating data.
        """

        enum_data_files = enum_keys(self.data_files)
        self.issue_instances = pd.DataFrame()
        self.full_issue_df = pd.DataFrame(
            columns=[
                "tables_affected",
                "columns_affected",
                "ROW_ID",
                "ERROR_ID",
                "rule_code",
                "rule_description",
                "rule_type",
                "la_level",
                "LAchildID",
            ]
        )
        self.rules_passed: list[str] = []

        self.rules_broken: list[str] = []
        self.rule_messages: list[str] = []
        self.la_rules_broken: list[str] = []

        registry = self.ruleset_registry

        rules_to_run = self.get_rules_to_run(registry, selected_rules)
        for rule in rules_to_run:
            data_files = copy.deepcopy(enum_data_files)
            ctx = RuleContext(rule)
            try:
                rule.func(data_files, ctx)
            except Exception as e:
                print(f"Error with rule {rule.code}: {type(e).__name__}, {e}")
            self.process_issues(rule, ctx)

        # df of all broken rule codes and related error messages.
        child_level_rules = pd.DataFrame(
            {"Rule code": self.rules_broken, "Rule Message": self.rule_messages}
        )
        # self.la_rules_broken is a list of issue_dfs, one per la-level rule that failed.
        try:
            self.la_rule_issues = pd.concat(self.la_rules_broken)
        except:
            # if la_rules_broken is still an empty list
            self.la_rule_issues = pd.DataFrame()



"""
Rule number: '8915'
Module: Child protection plans
Rule details: If <PersonDeathDate> (N00108) is present, then <CPPstartDate> (N00105) must be on or before <PersonDeathDate> (N00108)
Rule message: Child Protection Plan shown as starting after the child’s Date of Death

"""
from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

ChildProtectionPlans = CINTable.ChildProtectionPlans
CPPstartDate = ChildProtectionPlans.CPPstartDate
CPP_LAID = ChildProtectionPlans.LAchildID

ChildIdentifiers = CINTable.ChildIdentifiers
PersonDeathDate = ChildIdentifiers.PersonDeathDate
CI_LAID = ChildIdentifiers.LAchildID


# define characteristics of rule
@rule_definition(
    code="8915",
    module=CINTable.ChildProtectionPlans,
    message="Child Protection Plan shown as starting after the child’s Date of Death",
    affected_fields=[
        CPPstartDate,
        PersonDeathDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_CPP = data_container[ChildProtectionPlans].copy()
    df_CI = data_container[ChildIdentifiers].copy()

    df_CPP.index.name = "ROW_ID"
    df_CI.index.name = "ROW_ID"

    df_CPP.reset_index(inplace=True)
    df_CI.reset_index(inplace=True)

    # Remove rows with no death date

    df_CI = df_CI[df_CI[PersonDeathDate].notna()]

    # <CPPstartDate> (N00105) must be on or before <PersonDeathDate>

    # Join 2 tables together

    df = df_CPP.merge(
        df_CI,
        left_on=["LAchildID"],
        right_on=["LAchildID"],
        how="left",
        suffixes=("_CPP", "_CI"),
    )

    # Return those where dates don't align
    df = df[df[CPPstartDate] > df[PersonDeathDate]].reset_index()

    df["ERROR_ID"] = tuple(zip(df["LAchildID"], df[CPPstartDate], df[PersonDeathDate]))

    df_CPP_issues = (
        df_CPP.merge(df, left_on="ROW_ID", right_on="ROW_ID_CPP")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    df_CI_issues = (
        df_CI.merge(df, left_on="ROW_ID", right_on="ROW_ID_CI")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=ChildProtectionPlans, columns=[CPPstartDate], row_df=df_CPP_issues
    )
    rule_context.push_type_2(
        table=ChildIdentifiers, columns=[PersonDeathDate], row_df=df_CI_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_CPP = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CPPstartDate": pd.NA,  # Pass, no CP plan
            },
            {
                "LAchildID": "child2",
                "CPPstartDate": "27/06/2002",  # Fail, after death
            },
            {
                "LAchildID": "child3",
                "CPPstartDate": "07/02/1999",  # Pass, prior to death
            },
            {
                "LAchildID": "child4",
                "CPPstartDate": "26/05/2000",  # Pass, no death
            },
            {
                "LAchildID": "child5",
                "CPPstartDate": "26/05/2001",  # Fail, after death
            },
        ]
    )

    sample_CI = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # Pass
                "PersonDeathDate": "26/05/2000",
            },
            {
                "LAchildID": "child2",  # Fail
                "PersonDeathDate": "26/05/2000",
            },
            {
                "LAchildID": "child3",  # Pass
                "PersonDeathDate": "26/05/2000",
            },
            {
                "LAchildID": "child4",  # Pass
                "PersonDeathDate": pd.NA,
            },
            {
                "LAchildID": "child5",  # Fail
                "PersonDeathDate": "27/05/2000",
            },
        ]
    )

    sample_CPP[CPPstartDate] = pd.to_datetime(
        sample_CPP[CPPstartDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_CI[PersonDeathDate] = pd.to_datetime(
        sample_CI[PersonDeathDate], format="%d/%m/%Y", errors="coerce"
    )

    # Run rule function passing in our sample data
    result = run_rule(
        validate,
        {
            ChildProtectionPlans: sample_CPP,
            ChildIdentifiers: sample_CI,
        },
    )

    # The result contains a list of issues encountered
    issues_list = result.type2_issues
    assert len(issues_list) == 2

    issues = issues_list[1]

    issue_table = issues.table
    assert issue_table == ChildIdentifiers

    # check that the right columns were returned. Replace PersonDeathDate  with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [PersonDeathDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 2
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child2",  # ChildID
                    # Start date
                    pd.to_datetime("27/06/2002", format="%d/%m/%Y", errors="coerce"),
                    # Referral date
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child5",  # ChildID
                    # Start Date
                    pd.to_datetime("26/05/2001", format="%d/%m/%Y", errors="coerce"),
                    # Referral date
                    pd.to_datetime("27/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [4],
            },
        ]
    )

    assert issue_rows.equals(expected_df)

    assert result.definition.code == "8915"
    assert (
        result.definition.message
        == "Child Protection Plan shown as starting after the child’s Date of Death"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

CINdetails = CINTable.CINdetails
Section47 = CINTable.Section47

LAchildID = CINdetails.LAchildID
CINdetailsID = CINdetails.CINdetailsID
DateOfInitialCPC = Section47.DateOfInitialCPC
DateOfInitialCPC = CINdetails.DateOfInitialCPC


@rule_definition(
    code="2884",
    module=CINTable.Section47,
    message="An initial child protection conference is recorded at both the S47 and CIN Details level and it should only be recorded in one",
    affected_fields=[
        DateOfInitialCPC,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_47 = data_container[Section47].copy()
    df_cin = data_container[CINdetails].copy()

    df_47.index.name = "ROW_ID"
    df_cin.index.name = "ROW_ID"

    df_47.reset_index(inplace=True)
    df_cin.reset_index(inplace=True)

    merged_df = df_cin.merge(
        df_47,
        on=[LAchildID],
        suffixes=["_cin", "_47"],
    )

    condition = merged_df["DateOfInitialCPC_cin"] == merged_df["DateOfInitialCPC_47"]
    merged_df = merged_df[condition].reset_index()

    merged_df["ERROR_ID"] = tuple(
        zip(
            merged_df[LAchildID],
            merged_df["CINdetailsID_47"],
            merged_df["DateOfInitialCPC_cin"],
        )
    )

    df_47_issues = (
        df_47.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_47")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    df_cin_issues = (
        df_cin.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=Section47, columns=[DateOfInitialCPC], row_df=df_47_issues
    )
    rule_context.push_type_2(
        table=CINdetails, columns=[DateOfInitialCPC], row_df=df_cin_issues
    )


def test_validate():
    sample_section47 = pd.DataFrame(
        [
            {  # 0 fail: datecpc == (child1, cinID2)'s datecpc in cindetails table
                "LAchildID": "child1",
                "DateOfInitialCPC": "26/05/2000",
                "CINdetailsID": "cinID1",
            },
            {  # 1 fail: datecpc == (child1, cinID2)'s datecpc in cindetails table
                "LAchildID": "child1",
                "DateOfInitialCPC": "26/05/2000",
                "CINdetailsID": "cinID2",
            },
            {  # 2
                "LAchildID": "child2",
                "DateOfInitialCPC": "30/05/2000",
                "CINdetailsID": "cinID1",
            },
            {  # 3
                "LAchildID": "child3",
                "DateOfInitialCPC": "27/05/2000",
                "CINdetailsID": "cinID1",
            },
            {  # 4 fail: datecpc == (child3, cinID2)'s datecpc in cindetails table
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2000",
                "CINdetailsID": "cinID2",
            },
            {  # 5 fail: datecpc == (child3, cinID2)'s datecpc in cindetails table
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2000",
                "CINdetailsID": "cinID3",
            },
            {  # 6
                "LAchildID": "child3",
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID4",
            },
        ]
    )
    sample_cin_details = pd.DataFrame(
        [
            {  # 0
                "LAchildID": "child1",
                "DateOfInitialCPC": "26/10/1999",
                "CINdetailsID": "cinID1",
            },
            {  # 1 fail
                "LAchildID": "child1",
                "DateOfInitialCPC": "26/05/2000",
                "CINdetailsID": "cinID2",
            },
            {  # 2
                "LAchildID": "child2",
                "DateOfInitialCPC": "26/05/2000",
                "CINdetailsID": "cinID1",
            },
            {  # 3
                "LAchildID": "child3",
                "DateOfInitialCPC": "28/05/2000",
                "CINdetailsID": "cinID1",
            },
            {  # 4 fail
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2000",
                "CINdetailsID": "cinID2",
            },
            {  # 5
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2003",
                "CINdetailsID": "cinID3",
            },
            {  # 6
                "LAchildID": "child3",
                "DateOfInitialCPC": "14/03/2001",
                "CINdetailsID": "cinID4",
            },
        ]
    )
    sample_section47["DateOfInitialCPC"] = pd.to_datetime(
        sample_section47["DateOfInitialCPC"], format="%d/%m/%Y", errors="coerce"
    )
    sample_cin_details["DateOfInitialCPC"] = pd.to_datetime(
        sample_cin_details["DateOfInitialCPC"], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(
        validate,
        {
            Section47: sample_section47,
            CINdetails: sample_cin_details,
        },
    )

    issues_list = result.type2_issues
    assert len(issues_list) == 2
    issues = issues_list[0]

    issue_table = issues.table
    assert issue_table == Section47

    issue_columns = issues.columns
    assert issue_columns == [DateOfInitialCPC]

    issue_rows = issues.row_df
    assert len(issue_rows) == 4
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    "cinID1",
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child1",
                    "cinID2",
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child3",
                    "cinID2",
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [4],
            },
            {
                "ERROR_ID": (
                    "child3",
                    "cinID3",
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [5],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "2884"
    assert (
        result.definition.message
        == "An initial child protection conference is recorded at both the S47 and CIN Details level and it should only be recorded in one"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

Disabilities = CINTable.Disabilities
Disability = Disabilities.Disability


# define characteristics of rule
@rule_definition(
    code="8794",
    module=CINTable.Disabilities,
    message="Child has two or more disabilities with the same code",
    affected_fields=[Disability],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[Disabilities]

    # If there is more than one <Disability> (N00099) for a child, then none of the values should appear more than once
    df_orig = df.copy()
    df_orig.reset_index(inplace=True)

    # Create a 'counts' column, a count of rows partitioned by LAchildID and Disability such that
    # if an LAchildID-Disability combination appears twice, then it'll have a count of 2, and so on.
    df = df.groupby(["LAchildID", "Disability"]).size().reset_index(name="counts")

    # Add the count column back into the original dataframe, joining by LAchildID and Disability
    df = df.merge(df_orig, how="left", on=["LAchildID", "Disability"])

    # Hold all counts >= 2 (which are the error rows)
    df = df[df["counts"] >= 2]

    # Return original index for the error rows
    failing_indices = df.set_index("index").index

    rule_context.push_issue(table=Disabilities, field=Disability, row=failing_indices)


def test_validate():
    # 0     #1      #2      #3     #4      #5      #6      #7   #8     #9   #10   #11   #12
    ids = ["1", "1", "2", "3", "3", "4", "4", "5", "5", "6", "6", "6", "6"]
    dis_is = [
        "AAA",
        "AAA",
        "BBB",
        pd.NA,
        "MOTH",
        "AAAA",
        "AAAA",
        "AA",
        "BB",
        "AA",
        "AA",
        "CC",
        "CC",
    ]

    fake_dataframe = pd.DataFrame({"LAchildID": ids, "Disability": dis_is})

    result = run_rule(validate, {Disabilities: fake_dataframe})

    issues = list(result.issues)

    assert len(issues) == 8

    assert issues == [
        IssueLocator(CINTable.Disabilities, Disability, 0),
        IssueLocator(CINTable.Disabilities, Disability, 1),
        IssueLocator(CINTable.Disabilities, Disability, 5),
        IssueLocator(CINTable.Disabilities, Disability, 6),
        IssueLocator(CINTable.Disabilities, Disability, 9),
        IssueLocator(CINTable.Disabilities, Disability, 10),
        IssueLocator(CINTable.Disabilities, Disability, 11),
        IssueLocator(CINTable.Disabilities, Disability, 12),
    ]

    assert result.definition.code == "8794"
    assert (
        result.definition.message
        == "Child has two or more disabilities with the same code"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    RuleType,
    rule_definition,
)
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

ChildIdentifiers = CINTable.ChildIdentifiers
Header = CINTable.Header
PersonDeathDate = ChildIdentifiers.PersonDeathDate
ReferenceDate = Header.ReferenceDate


@rule_definition(
    code="8545Q",
    module=CINTable.ChildIdentifiers,
    rule_type=RuleType.QUERY,
    message="Please check and either amend data or provide a reason: Child's date of death should be within the census year",
    affected_fields=[PersonDeathDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildIdentifiers]
    df_ref = data_container[Header]

    # If present, <PersonDeathDate> (N00108) must be within [Period_of_Census]
    df = df[[PersonDeathDate]]

    # Death date must not be null, invalid text dates are made null in the line above
    df = df[df[PersonDeathDate].notna()]

    # Select out only the ReferenceDate column from the DataFrame
    ref_date_series = df_ref[ReferenceDate]
    collection_start, collection_end = make_census_period(ref_date_series)

    # DeathDate isn't in the financial year
    df = df[
        ~(
            (df[PersonDeathDate] >= collection_start)
            & (df[PersonDeathDate] <= collection_end)
        )
    ]

    failing_indices = df.index

    rule_context.push_issue(
        table=ChildIdentifiers, field=PersonDeathDate, row=failing_indices
    )


def test_validate():
    p_death = pd.to_datetime(
        [
            "2022-04-25",
            "2022-03-01",
            "2022-12-25",
            "2021-04-27",
            "2021-11-21",
            "2021-08-20",
            "2020-04-17",
            "1666-55-55",
            pd.NA,
        ],
        format="%Y/%m/%d",
        errors="coerce",
    )

    fake_ident = pd.DataFrame({PersonDeathDate: p_death})
    fake_head = pd.DataFrame({ReferenceDate: ["31/03/2022"]})

    result = run_rule(validate, {ChildIdentifiers: fake_ident, Header: fake_head})

    issues = list(result.issues)

    assert len(issues) == 3

    assert issues == [
        IssueLocator(CINTable.ChildIdentifiers, PersonDeathDate, 0),
        IssueLocator(CINTable.ChildIdentifiers, PersonDeathDate, 2),
        IssueLocator(CINTable.ChildIdentifiers, PersonDeathDate, 6),
    ]

    assert result.definition.code == "8545Q"
    assert (
        result.definition.message
        == "Please check and either amend data or provide a reason: Child's date of death should be within the census year"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

CINdetails = CINTable.CINdetails
PrimaryNeedCode = CINdetails.PrimaryNeedCode


# define characteristics of rule
@rule_definition(
    code="8650",
    module=CINTable.CINdetails,
    message="Primary Need Code invalid (see Primary Need table in CIN census code set)",
    affected_fields=[PrimaryNeedCode],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[CINdetails]

    # If present <PrimaryNeedCode> (N00101) must be a valid primary need code value
    PriNeed_list = ["N0", "N1", "N2", "N3", "N4", "N5", "N6", "N7", "N8", "N9"]

    # Primary Need Code is not in list.
    df = df[(~df["PrimaryNeedCode"].isin(PriNeed_list)) & df["PrimaryNeedCode"].notna()]

    failing_indices = df.index

    rule_context.push_issue(
        table=CINdetails, field=PrimaryNeedCode, row=failing_indices
    )


def test_validate():
    pri_need = ["N1", "N8", "AA", pd.NA, "N6", "BB", "N9"]

    fake_dataframe = pd.DataFrame({"PrimaryNeedCode": pri_need})

    result = run_rule(validate, {CINdetails: fake_dataframe})

    issues = list(result.issues)

    assert len(issues) == 2

    assert issues == [
        IssueLocator(CINTable.CINdetails, PrimaryNeedCode, 2),
        IssueLocator(CINTable.CINdetails, PrimaryNeedCode, 5),
    ]

    assert result.definition.code == "8650"
    assert (
        result.definition.message
        == "Primary Need Code invalid (see Primary Need table in CIN census code set)"
    )
